import sys
import numpy as np
from qtpy.QtWidgets import (
    QApplication, QMainWindow, QComboBox, QGridLayout, QLineEdit, QWidget, QFrame, QMessageBox
)
from qtpy.QtGui import QIcon
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from elasticipy.crystal_symmetries import SYMMETRIES
from elasticipy.gui.rotate_window import EulerBungeDialog
from elasticipy.gui.about import about
from elasticipy.tensors.elasticity import StiffnessTensor
from importlib import resources

from scipy.spatial.transform import Rotation

WHICH_OPTIONS = {'Mean': 'mean', 'Max': 'max', 'Min': 'min', 'Std. dev.': 'std'}


class ElasticityGUI(QMainWindow):
    def __init__(self):
        super().__init__()
        self.C_stiff = None
        self.coefficient_fields = {}
        self.setWindowTitle("elasticipy - GUI")
        self.initUI()
        self.C_invar= np.zeros(7)

    def selected_symmetry(self):
        symmetry_name = self.symmetry_selector.currentText()
        symmetry = SYMMETRIES[symmetry_name]
        if symmetry_name == "Trigonal" or symmetry_name == "Tetragonal":
            space_group_text = self.point_group_selector.currentText()
            try:
                symmetry = symmetry[space_group_text]
            except KeyError:
                # If no SG is selected, just choose one
                symmetry = list(symmetry.values())[0]
        elif symmetry_name == "Monoclinic":
            diad_index = self.diag_selector.currentText()
            symmetry = symmetry[diad_index]
        return symmetry

    def initUI(self):
        # Main layout
        main_layout = QVBoxLayout()

        #######################################################################################
        # Material's symmetry and other parameters
        #######################################################################################
        selectors_layout = QHBoxLayout()

        # Symmetry selection
        self.symmetry_selector = QComboBox()
        self.symmetry_selector.addItems(SYMMETRIES.keys())
        self.symmetry_selector.currentIndexChanged.connect(self.update_fields)
        selectors_layout.addWidget(QLabel("Crystal symmetry:"))
        selectors_layout.addWidget(self.symmetry_selector)

        # Space Group selection
        self.point_group_selector = QComboBox()
        self.point_group_selector.setMinimumWidth(140)
        self.point_group_selector.addItems(['', ''])
        self.point_group_selector.currentIndexChanged.connect(self.update_fields)
        selectors_layout.addWidget(QLabel("Point group:"))
        selectors_layout.addWidget(self.point_group_selector)

        # Diad selection
        self.diag_selector = QComboBox()
        self.diag_selector.addItems(SYMMETRIES['Monoclinic'].keys())
        self.diag_selector.currentIndexChanged.connect(self.update_fields)
        selectors_layout.addWidget(QLabel("Diad convention:"))
        selectors_layout.addWidget(self.diag_selector)

        # About button
        self.about_button = QPushButton("About")
        self.about_button.setFixedHeight(24)
        self.about_button.setMaximumWidth(60)
        self.about_button.clicked.connect(self.show_about)
        selectors_layout.addStretch()
        selectors_layout.addWidget(self.about_button)

        # Add horizontal separator
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)

        # Add selectors_layout to main layout
        main_layout.addLayout(selectors_layout)
        main_layout.addWidget(separator)

        #######################################################################################
        # Matrix components
        #######################################################################################
        grid = QGridLayout()
        for i in range(6):
            for j in range(i, 6):
                field = QLineEdit()
                field.setPlaceholderText(f"C{i+1}{j+1}")
                self.coefficient_fields[(i, j)] = field
                field.textChanged.connect(self.update_dependent_fields)
                grid.addWidget(field, i, j)
        main_layout.addLayout(grid)

        #######################################################################################
        # Bottom panel
        #######################################################################################
        bottom_layout = QHBoxLayout()

        ############################################
        # Plotting options
        ############################################
        left_panel_layout = QVBoxLayout()

        # E, G or nu selector
        parameter_layout = QHBoxLayout()
        self.plotting_selector = QComboBox()
        self.plotting_selector.addItems(['Young modulus', 'Shear modulus', 'Poisson ratio', 'Linear compressibility'])
        self.plotting_selector.currentIndexChanged.connect(self.update_plotting_selectors)
        parameter_layout.addWidget(QLabel("Parameter:"))
        parameter_layout.addWidget(self.plotting_selector)
        left_panel_layout.addLayout(parameter_layout)

        # Plotting style
        style_layout = QHBoxLayout()
        self.plot_style_selector = QComboBox()
        self.plot_style_selector.addItems(['3D', 'Sections', 'Pole Figure'])
        self.plot_style_selector.currentIndexChanged.connect(self.update_plotting_selectors)
        style_layout.addWidget(QLabel("Plot type:"))
        style_layout.addWidget(self.plot_style_selector)
        left_panel_layout.addLayout(style_layout)

        # 'which' selector
        which_layout = QHBoxLayout()
        self.which_selector = QComboBox()
        self.which_selector.addItems(WHICH_OPTIONS.keys())
        which_layout.addWidget(QLabel("Value:"))
        which_layout.addWidget(self.which_selector)
        left_panel_layout.addLayout(which_layout)

        # Plot button
        self.calculate_button = QPushButton("Plot")
        self.calculate_button.setEnabled(False)
        self.calculate_button.clicked.connect(self.calculate_and_plot)
        left_panel_layout.addWidget(self.calculate_button)

        self.euler_button = QPushButton("Apply rotation")
        self.euler_button.setEnabled(False)
        self.euler_button.setToolTip("Rotate stiffness tensor (Bunge ZXZ)")
        self.euler_button.clicked.connect(self.open_euler_dialog)
        left_panel_layout.addWidget(self.euler_button)

        # Add horizontal separator
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)
        left_panel_layout.addWidget(separator)

        ############################################
        # Numeric results
        ############################################
        self.result_labels = {}
        RESULT_GROUPS = {
            "Young modulus": [
                ("E_mean", "Mean"),
                ("E_voigt", "Voigt"),
                ("E_reuss", "Reuss"),
                ("E_hill", "Hill"),
            ],
            "Shear modulus": [
                ("G_mean", "Mean"),
                ("G_voigt", "Voigt"),
                ("G_reuss", "Reuss"),
                ("G_hill", "Hill"),
            ],
            "Poisson ratio": [
                ("nu_mean", "Mean"),
                ("nu_voigt", "Voigt"),
                ("nu_reuss", "Reuss"),
                ("nu_hill", "Hill"),
            ],
            "Linear compressibility (x1000)": [
                ("Beta_mean", "Mean"),
                ("Beta_voigt", "Voigt"),
                ("Beta_reuss", "Reuss"),
                ("Beta_hill", "Hill"),
            ],
            "Other": [
                ("K", "Bulk modulus"),
                ("Z", "Zener ratio"),
                ("A", "Univ. anisotropy factor"),
            ]
        }

        ############################################
        # Numeric results (grouped)
        ############################################
        self.result_labels = {}
        for group_name, items in RESULT_GROUPS.items():

            # Group title
            group_label = QLabel(group_name + ":")
            group_label.setStyleSheet("font-weight: bold; margin-top: 6px;")
            left_panel_layout.addWidget(group_label)

            # Indented layout
            indent_layout = QVBoxLayout()
            indent_layout.setContentsMargins(15, 0, 0, 0)

            for key, label_text in items:
                row = QHBoxLayout()

                label_name = QLabel(f"{label_text}:")
                label_value = QLabel("—")
                label_value.setMinimumWidth(100)
                label_value.setStyleSheet("font-family: Consolas, Courier;")

                self.result_labels[key] = label_value

                row.addWidget(label_name)
                row.addStretch()
                row.addWidget(label_value)

                indent_layout.addLayout(row)

            left_panel_layout.addLayout(indent_layout)

        # Fill space
        left_panel_layout.addStretch()


        bottom_layout.addLayout(left_panel_layout,1)

        ############################################
        # Plotting area
        ############################################
        # Add separator between the top and the bottom panels
        separator2 = QFrame()
        separator2.setFrameShape(QFrame.HLine)
        separator2.setFrameShadow(QFrame.Sunken)
        main_layout.addWidget(separator2)

        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        bottom_layout.addWidget(self.canvas,4)

        #######################################################################################
        # Main widget
        #######################################################################################
        main_layout.addLayout(bottom_layout)
        central_widget = QWidget()
        central_widget.setLayout(main_layout)
        self.setCentralWidget(central_widget)

        #######################################################################################
        # Initialize at load
        #######################################################################################
        self.symmetry_selector.setCurrentText('Triclinic')
        self.plotting_selector.setCurrentText('Young modulus')
        self.which_selector.setEnabled(False)


    def update_fields(self):
        # Deactivate unused fields
        active_fields = self.selected_symmetry().required
        for (i, j), field in self.coefficient_fields.items():
            if (i, j) in active_fields:
                field.setEnabled(True)
                if field.text() == "0":
                    field.setText('')
            else:
                field.setEnabled(False)
                field.setText('0')

        # Turn on/off SG selection
        selected_symmetry_name = self.symmetry_selector.currentText()
        trig_or_tetra = selected_symmetry_name in ("Trigonal", "Tetragonal")
        self.point_group_selector.setEnabled(trig_or_tetra) # Turn on/off SG selector
        if trig_or_tetra:
            # Change list of possible SGs
            self.point_group_selector.setEnabled(True)
            space_groups = SYMMETRIES[selected_symmetry_name].keys()
            for i in range(len(space_groups)):
                self.point_group_selector.setItemText(i, list(space_groups)[i])
        self.diag_selector.setEnabled(selected_symmetry_name == "Monoclinic")

    def update_plotting_selectors(self):
        if (self.plotting_selector.currentText() == "Young modulus" or
            self.plotting_selector.currentText() == "Linear compressibility" or
                self.plot_style_selector.currentIndex() == 1):
            self.which_selector.setEnabled(False)
        else:
            self.which_selector.setEnabled(True)

    def calculate_and_plot(self):
        stiff = self.C_stiff
        self.figure.clear()
        requested_value = self.plotting_selector.currentText()
        if requested_value == "Young modulus":
            value = stiff.Young_modulus
            plot_kwargs = {}
        elif requested_value == 'Linear compressibility':
            value = stiff.linear_compressibility
            plot_kwargs = {}
        else:
            if requested_value == 'Shear modulus':
                value = stiff.shear_modulus
            else:
                value = stiff.Poisson_ratio
            plot_kwargs = {'which': WHICH_OPTIONS[self.which_selector.currentText()]}
        if self.plot_style_selector.currentIndex() == 0:
            value.plot3D(fig=self.figure, backend='matplotlib', **plot_kwargs)
        elif self.plot_style_selector.currentIndex() == 1:
            value.plot_xyz_sections(fig=self.figure)
        else:
            value.plot_as_pole_figure(fig=self.figure, **plot_kwargs)
        self.canvas.draw()
        invariants = np.array(stiff.linear_invariants() + stiff.quadratic_invariants())
        if not np.all(np.isclose(invariants, self.C_invar)):
            self.C_invar = invariants
            self.result_labels["E_mean"].setText(f"{stiff.Young_modulus.mean():.3f}")
            self.result_labels["G_mean"].setText(f"{stiff.shear_modulus.mean():.3f}")
            self.result_labels["nu_mean"].setText(f"{stiff.Poisson_ratio.mean():.3f}")
            self.result_labels["Beta_mean"].setText(f"{stiff.linear_compressibility.mean()*1000:.3f}")
            for method in ['voigt', 'reuss', 'hill']:
                C = stiff.average(method=method)
                self.result_labels[f"E_{method}"].setText(f"{C.Young_modulus.eval([1,0,0]):.3f}")
                self.result_labels[f"G_{method}"].setText(f"{C.shear_modulus.eval([1, 0, 0],[0,1,0]):.3f}")
                self.result_labels[f"nu_{method}"].setText(f"{C.Poisson_ratio.eval([1, 0, 0], [0, 1, 0]):.3f}")
                self.result_labels[f"Beta_{method}"].setText(f"{C.linear_compressibility.eval([1, 0, 0])*1000:.3f}")
            self.result_labels["K"].setText(f"{stiff.bulk_modulus:.3f}")
            try:
                Z = stiff.Zener_ratio()
                self.result_labels["Z"].setText(f"{stiff.Zener_ratio():.3f}")
            except ValueError:
                self.result_labels["Z"].setText("—")
            self.result_labels["A"].setText(f"{stiff.universal_anisotropy:.3f}")


    def update_dependent_fields(self):
        symmetry = self.selected_symmetry()
        for equality in symmetry.equal:
            try:
                ref_value = float(self.coefficient_fields[equality[0]].text())
                for index in equality[1]:
                        self.coefficient_fields[index].setText(f"{ref_value}")
            except ValueError:
                pass
        for opposite in symmetry.opposite:
            try:
                ref_value = float(self.coefficient_fields[opposite[0]].text())
                for index in opposite[1]:
                        self.coefficient_fields[index].setText(f"{-ref_value}")
            except ValueError:
                pass
        if symmetry.C11_C12:
            try:
                C11 = float(self.coefficient_fields[(0, 0)].text())
                C12 = float(self.coefficient_fields[(0, 1)].text())
                for index in symmetry.C11_C12:
                    self.coefficient_fields[index].setText(f"{0.5*(C11-C12)}")
            except ValueError:
                pass
        coefficients = np.zeros((6, 6))
        for (i, j), field in self.coefficient_fields.items():
            try:
                coefficients[i, j] = float(field.text())
            except ValueError:
                coefficients[i, j] = 0
        Csym = coefficients + np.tril(coefficients.T, -1) # Rebuild the lower triangular part
        try:
            self.C_stiff = StiffnessTensor(Csym)
            self.calculate_button.setEnabled(True)
            self.euler_button.setToolTip("Plot directional dependence")
            self.euler_button.setEnabled(True)
            self.euler_button.setToolTip("Rotate stiffness tensor (Bunge ZXZ)")
        except ValueError:
            self.calculate_button.setEnabled(False)
            error_msg = "The stiffness tensor is not definite positive!"
            self.calculate_button.setToolTip(error_msg)
            self.euler_button.setEnabled(False)
            self.euler_button.setToolTip(error_msg)

    def show_about(self):
        dialog = QDialog(self)
        dialog = about(dialog)
        dialog.exec_()

    def open_euler_dialog(self):
        if not hasattr(self, "euler_dialog"):
            self.current_config = {'sym': self.symmetry_selector.currentText(),
                              'pg': self.point_group_selector.currentText(),
                              'diag': self.diag_selector.currentText()}
            self.euler_dialog = EulerBungeDialog(self, C=self.C_stiff.matrix())
            self.euler_dialog.anglesChanged.connect(self.update_from_euler)
        self.euler_dialog.show()

    def update_from_euler(self, phi1, Phi, phi2):
        C0 = StiffnessTensor(self.euler_dialog.C)
        rot = Rotation.from_euler(seq='ZXZ', angles=[phi1, Phi, phi2], degrees=True)
        if np.any(np.array([phi1, Phi, phi2])) != 0.:
            self.symmetry_selector.setCurrentText('Triclinic')
            C_new = C0.rotate(rot).matrix()
            for (i, j), field in self.coefficient_fields.items():
                self.coefficient_fields[(i,j)].setText(f"{C_new[i,j]}")
        else:
            for (i, j), field in self.coefficient_fields.items():
                self.coefficient_fields[(i,j)].setText(f"{self.euler_dialog.C[i,j]}")
            self.symmetry_selector.setCurrentText(self.current_config['sym'])
            self.point_group_selector.setCurrentText(self.current_config['pg'])
            self.diag_selector.setCurrentText(self.current_config['diag'])
        if self.euler_dialog.live_button.isChecked():
            self.calculate_and_plot()

def crystal_elastic_plotter():
    app = QApplication(sys.argv)
    try:
        path_to_icon = str(resources.files("elasticipy.resources") / "favicon.png")
        icon = QIcon(str(path_to_icon))
    except Exception:
        icon = QIcon()
    app.setWindowIcon(icon)
    window = ElasticityGUI()
    window.setWindowIcon(icon)
    window.show()
    sys.exit(app.exec_())


from qtpy.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout,
    QLabel, QPushButton
)

if __name__ == "__main__":
    crystal_elastic_plotter()


import unittest
from elasticipy.yield_criteria import DruckerPrager, MohrCoulomb, TrescaCriterion, VonMisesCriterion
from elasticipy.tensors.stress_strain import StressTensor, StrainTensor
import numpy as np
from pytest import approx

tensile_x = StressTensor.tensile([1, 0, 0], 1)
tensile_y = StressTensor.tensile([0, 1, 0], 1)
shear = StressTensor.shear([1, 0, 0], [0, 1, 0], 1)
biaxial = tensile_x + 2 *tensile_y
K = 3 / 2 * 1 / 3 ** 0.5

def generic_test_from_tensile_compression_stresses(obj, criterion):
    with obj.assertRaises(ValueError) as context:
        criterion.from_tensile_compression_stress(100, 100)
    obj.assertEqual(str(context.exception), 'The compression yield stress must be negative.')
    with obj.assertRaises(ValueError) as context:
        criterion.from_tensile_compression_stress(-100, -150)
    obj.assertEqual(str(context.exception), 'The tensile yield stress must be positive.')

class TestDruckerPrager(unittest.TestCase):
    def test_vonMises(self):
        pg_vm = DruckerPrager(0, 1)
        for stress in [tensile_x, tensile_y, shear, biaxial]:
            assert pg_vm.yield_function(stress) == stress.vonMises() / np.sqrt(3) - 1

            mises_crit = VonMisesCriterion()
            np.testing.assert_array_almost_equal(pg_vm.normal(stress).matrix, mises_crit.normal(stress).matrix)

    def test_yield_function(self):
        k, alpha = 2, 0.1
        pg = DruckerPrager(alpha, k)
        sy_tensile = k/(alpha + 3**(-0.5))
        sy_compres = k / (alpha - 3 ** (-0.5))
        assert pg.yield_function(tensile_x * sy_tensile) == 0.0
        assert pg.yield_function(tensile_x * sy_compres) == 0.0

    def test_inequality_fit(self):
        c, phi = 2, -10
        dp_1 = DruckerPrager.from_cohesion_friction_angle(c, phi, fit='inside')
        dp_2 = DruckerPrager.from_cohesion_friction_angle(c, phi, fit='middle')
        dp_3 = DruckerPrager.from_cohesion_friction_angle(c, phi, fit='outside')
        rand_stress = StressTensor.rand(shape=1000, seed=123)
        f1 = dp_1.yield_function(rand_stress)
        f2 = dp_2.yield_function(rand_stress)
        f3 = dp_3.yield_function(rand_stress)
        assert np.all(np.logical_and(f1 > f2, f2 > f3))

    def test_normality(self):
        k, alpha = 2, 0.1
        dp = DruckerPrager(alpha, k)
        stress = StressTensor.eye()
        unit_strain = StrainTensor.eye()
        assert dp.normal(stress) == unit_strain/unit_strain.eq_strain()

    def test_from_tensile_compression_stress(self):
        s_t = 100
        s_c = -150
        mc = DruckerPrager.from_tensile_compression_stress(s_t, s_c)
        assert mc.yield_function(tensile_x * s_t) == approx(0.)
        assert mc.yield_function(tensile_x * s_c) == approx(0.)

    def test_from_tensile_compression_stress_valid_args(self):
        generic_test_from_tensile_compression_stresses(self, DruckerPrager)


class TestMohrCoulomb(unittest.TestCase):
    def test_Tresca(self):
        mv_tr = MohrCoulomb(1, 0)
        for stress in [tensile_x, tensile_y, shear, biaxial]:
            assert mv_tr.yield_function(stress) == stress.Tresca() - 2

            tresca_crit = TrescaCriterion()
            np.testing.assert_array_almost_equal(mv_tr.normal(stress).matrix, tresca_crit.normal(stress).matrix)

    def test_normality(self):
        mv_tr = MohrCoulomb(1, 0)
        stress = StressTensor.eye()
        unit_strain = StrainTensor.eye()
        assert mv_tr.normal(stress) == unit_strain/unit_strain.eq_strain()

    def test_from_tensile_compression_stress(self):
        s_t = 100
        s_c = -150
        mc = MohrCoulomb.from_tensile_compression_stress(s_t, s_c)
        assert mc.yield_function(tensile_x * s_t) == 0.
        assert mc.yield_function(tensile_x * s_c) == 0.

    def test_from_tensile_compression_stress_valid_args(self):
        generic_test_from_tensile_compression_stresses(self, MohrCoulomb)

class TestTrescaCriterion(unittest.TestCase):
    def test_normality_Tresca(self):
        biaxial = (StressTensor.tensile([1,0,0],[0, 1, 1, 1, 1, 1, 0]) +
                   StressTensor.tensile([0,1,0],[-1, -1, -0.5, 0, 0.5, 1, 1]))
        n = TrescaCriterion.normal(biaxial)
        assert n[0] == VonMisesCriterion.normal(biaxial[0])
        assert n[2] == K * np.diag([1, -1, 0])
        assert n[2] == K * np.diag([1, -1, 0])
        assert n[3] == VonMisesCriterion.normal(biaxial[3])
        assert n[4] == K * np.diag([1, 0, -1])
        assert n[5] == VonMisesCriterion.normal(biaxial[5])
        assert n[6] == VonMisesCriterion.normal(biaxial[6])

        # Check that the magnitude of the normal is 1
        np.testing.assert_array_equal(n.eq_strain(), np.ones(biaxial.shape))
        triaxial = StressTensor(np.diag([1,2,4]))
        n = TrescaCriterion.normal(triaxial)
        assert n == K * np.diag([-1, 0, 1])
        assert n.eq_strain() == 1.0

class TestVonMisesCriterion(unittest.TestCase):
    def test_scale_stress(self):
        vm = VonMisesCriterion(yield_stress=100)
        stress = vm.scale_stress_to_yield_surface(tensile_x)
        np.testing.assert_array_almost_equal(stress.matrix, tensile_x.matrix * 100)

    def test_normality_J2(self):
        tensile_stress = StressTensor.tensile([1, 0, 0], 1)
        normal = VonMisesCriterion.normal(tensile_stress)
        assert normal == np.diag([1., -0.5, -0.5])

        shear_stress = StressTensor.shear([1, 0, 0], [0, 1, 0], 1)
        normal = VonMisesCriterion.normal(shear_stress)
        normal_th = K * np.array([[0, 1, 0],
                                  [1, 0, 0],
                                  [0, 0, 0]])
        np.testing.assert_array_almost_equal(normal.matrix, normal_th)

if __name__ == '__main__':
    unittest.main()



"""
===========================================================
Plot directional dependence of elastic moduli
===========================================================

This example shows different ways to plot the directional dependence of elastic moduli,
such as Young's modulus, Poisson's ratio, and shear modulus.
"""

from elasticipy.tensors.elasticity import StiffnessTensor
from matplotlib import pyplot as plt

C = StiffnessTensor.orthorhombic(phase_name='forsterite',
                                C11=320, C12=68.2, C13=71.6,
                                C22=196.5, C23=76.8, C33=233.5, C44=64, C55=77, C66=78.7)
rho = 3.355

cp, cs_fast, cs_slow = C.wave_velocity(rho)
fig = plt.figure(figsize=(12,4))
cp.plot_as_pole_figure(subplot_args=(131,), title='p wave', fig=fig)
cs_fast.plot_as_pole_figure(subplot_args=(132,), title='s wave (fast)', fig=fig)
cs_slow.plot_as_pole_figure(subplot_args=(133,), title='s wave (slow)', fig=fig)
fig.show()


"""
============================================================
3D plot the Drucker-Prager and Mohr-Coulomb yield surface
============================================================

We plot the Drucker-Prager and Mohr-coulomb yield surfaces in the 3D principal stresses space.

.. note::

    This example is inspired by the figure shown on wikipedia to illustrate the DP yield criterion (see
    `here <https://en.wikipedia.org/wiki/Drucker%E2%80%93Prager_yield_criterion>`_).
"""
from elasticipy.yield_criteria import DruckerPrager, MohrCoulomb

pg = DruckerPrager.from_cohesion_friction_angle(2, -20)
fig=pg.plot3D(xrange=(-6, 6), yrange=(-6, 6), zrange=(-6, 6))

mc = MohrCoulomb(2, -20)
mc.plot3D(fig=fig, color='blue')


import unittest

from orix.quaternion import Orientation

from elasticipy.crystal_texture import DiscreteTexture, FibreTexture, CompositeTexture, UniformTexture
from orix.vector import Miller, Vector3d
from orix.crystal_map import Phase
import numpy as np
from elasticipy.tensors.elasticity import StiffnessTensor
from pytest import approx
from scipy.stats import kstest

PHASE = Phase(point_group="m-3m")
C = StiffnessTensor.monoclinic(phase_name='TiNi',
                               C11=231, C12=127, C13=104,
                               C22=240, C23=131, C33=175,
                               C44=81, C55=11, C66=85,
                               C15=-18, C25=1, C35=-3, C46=3)

def orientation_checker(texture, hkl, uvw):
    orientation = texture.orientation
    m_hkl = Miller(uvw=hkl, phase=PHASE)
    x = ~orientation * m_hkl.symmetrise(unique=True)
    u_hkl = x.data
    cos_hkl = np.dot(u_hkl, [0, 0, 1]) / np.linalg.norm(u_hkl, axis=1)
    assert np.any(np.isclose(cos_hkl, 1))

    m_uvw = Miller(uvw=uvw, phase=PHASE)
    x = ~orientation * m_uvw.symmetrise(unique=True)
    u_uvw = x.data
    cos_uvw = np.dot(u_uvw, [1, 0, 0]) / np.linalg.norm(u_uvw, axis=1)
    assert np.any(np.isclose(cos_uvw, 1))


class TestUniformTexture(unittest.TestCase):
    def test_uniform(self):
        t = UniformTexture()
        assert C * t == C.Voigt_average()
        S = C.inv()
        assert S * t == S.Reuss_average()

    def test_uniform_random(self):
        t = UniformTexture()
        o = t.sample(10000, seed=132)
        v = Vector3d(~o * Vector3d([0,0,1]))
        ks1 = kstest(v.data[:, 2], 'uniform', args=(-1, 2))
        ks2 = kstest(v.azimuth, 'uniform', args=(0, 2*np.pi))
        assert 0.05 < ks1[1] < 0.95 # Use p-value
        assert 0.05 < ks2[1] < 0.95  # Use p-value

class TestDiscreteTexture(unittest.TestCase):
    def test_cube(self):
        t = DiscreteTexture.cube()
        orientation_checker(t, [0,0,1], [1,0,0])

    def test_Goss(self):
        t = DiscreteTexture.Goss()
        orientation_checker(t, [1,1,0], [1,0,0])

    def test_Brass(self):
        t = DiscreteTexture.brass()
        orientation_checker(t, [1,1,0], [1,1,2])

    def test_GossBrass(self):
        t = DiscreteTexture.GossBrass()
        orientation_checker(t, [1,1,0], [1,1,5])

    def test_Copper(self):
        t = DiscreteTexture.copper()
        orientation_checker(t, [1,1,2], [1,1,1])

    def test_A(self):
        t = DiscreteTexture.A()
        orientation_checker(t, [1,1,0], [1,1,1])

    def test_P(self):
        t = DiscreteTexture.P()
        orientation_checker(t, [0,1,1], [2,1,1])

    def test_CuT(self):
        t = DiscreteTexture.CuT()
        orientation_checker(t, [5,5,2], [1,1,5])

    def test_S(self):
        t = DiscreteTexture.S()
        orientation_checker(t, [1,2,3], [6,3,4])

    def test_repr(self):
        t = DiscreteTexture.Goss()
        assert t.__repr__() == 'Crystallographic texture\nφ1=0.00°, ϕ=45.00°, φ2=0.00°'

    def test_mult_stiffness(self):
        t = DiscreteTexture.S()
        Crot = C * t
        assert Crot == C * t.orientation

    def test_mult_float(self):
        t = DiscreteTexture.Goss()
        tm = t * 0.5
        assert isinstance(tm, DiscreteTexture)
        assert t.orientation == tm.orientation
        assert tm.weight == 0.5

        tm2 = 0.5 * t
        assert isinstance(tm2, DiscreteTexture)
        assert t.orientation == tm2.orientation

    def test_discrete_random(self):
        t = DiscreteTexture.Goss()
        o = t.sample(10)
        for oi in o:
            assert oi == t.orientation

class TestFibreTexture(unittest.TestCase):
    def test_from_Euler(self):
        t = FibreTexture.from_Euler(phi1=0, Phi=10)
        assert t.__repr__() == 'Fibre texture\nφ1= 0°, ϕ= 10°'

        t = FibreTexture.from_Euler(phi1=0, phi2=10)
        assert t.__repr__() == 'Fibre texture\nφ1= 0°, φ2= 10°'

        t = FibreTexture.from_Euler(Phi=0, phi2=10)
        assert t.__repr__() == 'Fibre texture\nϕ= 0°, φ2= 10°'

        t = FibreTexture.from_Euler(Phi=0, phi2=10 * np.pi / 180, degrees=False)
        assert t.__repr__() == 'Fibre texture\nϕ= 0.0°, φ2= 10.0°'

    def test_from_Euler_axis(self):
        t = FibreTexture.from_Euler(Phi=0, phi2=0)
        assert t.axis.dot(Vector3d([0,0,1])) == 1. or t.axis.dot(Vector3d([0,0,1])) == -1.

        t = FibreTexture.from_Euler(phi1=0, phi2=0)
        assert t.axis.dot(Vector3d([0,0,1])) == 1. or t.axis.dot(Vector3d([1,0,0])) == -1.

    def test_from_Miller_axis(self):
        m = Miller(uvw=[1, 0, 0], phase=PHASE)
        texture = FibreTexture.from_Miller_axis(m, [0, 0, 1])
        assert texture.__repr__() == 'Fibre texture\n<1. 0. 0.> || [0, 0, 1]'

    def test_mult_stiffness(self):
        # Check that a fibre texture along [0,0,1] leads to transverse isotropy
        m = Miller(uvw=[1, 0, 0], phase=PHASE)
        texture = FibreTexture.from_Miller_axis(m, [0, 0, 1])
        Crot = C * texture
        Exy = Crot.Young_modulus.eval([[1,0,0],[1,1,0],[0,1,0]])
        assert (approx(Exy[0]) == Exy[1]) and (approx(Exy[0]) == Exy[2])

    def test_mult_float(self):
        m = Miller(uvw=[1, 0, 0], phase=PHASE)
        t = FibreTexture.from_Miller_axis(m, [0, 0, 1])
        tm = t * 0.5
        assert isinstance(tm, FibreTexture)
        assert t.orientation == tm.orientation
        assert tm.weight == 0.5

        tm2 = 0.5 * t
        assert isinstance(tm2, FibreTexture)
        assert t.orientation == tm2.orientation

    def test_fibre_random(self):
        t = FibreTexture.from_Euler(Phi=0, phi2=0)
        o = t.sample(1000, seed=123)
        v = Vector3d(~o * Vector3d([1,0,0]))
        ks = kstest(v.azimuth, 'uniform', args=(0, 2*np.pi))
        assert 0.05 < ks[1] < 0.95  # Use p-value
        v = Vector3d(~o * Vector3d([0,0,1]))
        np.testing.assert_array_almost_equal(v.data[:,0], np.zeros(1000))
        np.testing.assert_array_almost_equal(v.data[:, 1], np.zeros(1000))

        m = Miller(uvw=[1, 1, 1], phase=PHASE)
        t = FibreTexture.from_Miller_axis(m, [0, 0, 1])
        o = t.sample(1000)
        v = Vector3d(~o * m)
        np.testing.assert_array_almost_equal(v.data[:,0], np.zeros(1000))
        np.testing.assert_array_almost_equal(v.data[:, 1], np.zeros(1000))

    def test_alpha(self):
        alpha = FibreTexture.alpha()
        o = alpha.sample()
        phi1, Phi, phi2 = o.in_euler_fundamental_region().T
        assert np.allclose(np.mod(phi1, np.pi), 0.)
        assert np.allclose(np.mod(phi2, np.pi/2), np.pi/4)

    def test_gamma(self):
        gamma = FibreTexture.gamma()
        o = gamma.sample()
        phi1, Phi, phi2 = o.in_euler_fundamental_region().T
        assert np.allclose(np.mod(Phi, np.pi/2), np.radians(54.73561032))
        assert np.allclose(np.mod(phi2, np.pi/2), np.pi/4)

    def test_epsilon(self):
        epsilon = FibreTexture.epsilon()
        o = epsilon.sample()
        phi1, Phi, phi2 = o.in_euler_fundamental_region().T
        assert np.allclose(np.mod(phi1, np.pi), np.pi/2)
        assert np.allclose(np.mod(phi2, np.pi/2), np.pi/4)

class TestCompositeTexture(unittest.TestCase):
    def test_add_Textures(self):
        tg = DiscreteTexture.Goss()
        tb = DiscreteTexture.brass()
        tm = tb + tg
        assert isinstance(tm, CompositeTexture)
        wgts = [t.weight for t in tm.texture_list]
        assert wgts == [1., 1.]
        assert tm.texture_list[0] == tb
        assert tm.texture_list[1] == tg

    def test_mult(self):
        tg = DiscreteTexture.Goss()
        tb = DiscreteTexture.brass()
        tm = 0.5 * (tb + tg)
        wgts = [t.weight for t in tm.texture_list]
        assert wgts == [0.5, 0.5]

    def test_add_TexturesMix(self):
        tg = DiscreteTexture.Goss()
        tb = DiscreteTexture.brass()
        tc = DiscreteTexture.cube()
        tm1 = 0.5 * tg + 0.3 * tb
        tm2 = tm1 + 0.4 * tc
        assert isinstance(tm2, CompositeTexture)
        orientations = [t.orientation for t in tm2.texture_list]
        assert orientations == [tg.orientation, tb.orientation, tc.orientation]
        weights = [t.weight for t in tm2.texture_list]
        assert weights == [0.5, 0.3, 0.4]

        tm3 = 0.4 * tc + tm1
        assert isinstance(tm3, CompositeTexture)
        orientations = [t.orientation for t in tm3.texture_list]
        assert orientations == [tc.orientation, tg.orientation, tb.orientation]
        weights = [t.weight for t in tm3.texture_list]
        assert weights == [0.4, 0.5, 0.3]

        tm4 = tm1 + tm2
        assert isinstance(tm4, CompositeTexture)
        orientations = [t.orientation for t in tm4.texture_list]
        assert orientations == [t.orientation for t in tm1.texture_list] + [t.orientation for t in tm2.texture_list]

    def test_repr(self):
        t1 = DiscreteTexture.Goss()
        m = Miller(uvw=[1, 0, 0], phase=PHASE)
        t2 = FibreTexture.from_Miller_axis(m, [0, 0, 1])
        t3 = FibreTexture.from_Euler(phi1=0, Phi=10)
        t4 = UniformTexture()
        tm = t1 + t2 + t3 + 0.5*t4
        assert isinstance(tm, CompositeTexture)
        expected_str = ('Mixture of crystallographic textures\n'
                        ' Wgt.  Type      Component\n'
                        ' ------------------------------------------------------------\n')
        expected_str += ' 1.00  discrete  φ1=0.00°, ϕ=45.00°, φ2=0.00°\n'
        expected_str += ' 1.00  fibre     <1. 0. 0.> || [0, 0, 1]\n'
        expected_str += ' 1.00  fibre     φ1= 0°, ϕ= 10°\n'
        expected_str += ' 0.50  uniform   Uniform distribution over SO(3)'
        assert tm.__repr__() == expected_str

    def test_mean(self):
        t2 = DiscreteTexture.cube()
        t1 = DiscreteTexture.Goss()
        tm = t1 + t2
        Cmean = C * tm
        assert isinstance(Cmean, StiffnessTensor)
        ori = Orientation.from_euler([[0, 0, 0], [0, 45, 0]], degrees=True)
        assert Cmean == (C * ori).Voigt_average()
        assert Cmean == tm * C

        tm_wgt = 1.5 * t1 + 0.5 * t2
        Cmean = tm_wgt.mean_tensor(C)
        assert isinstance(Cmean, StiffnessTensor)
        Cmean_voigt = StiffnessTensor.weighted_average((C * t1, C * t2), [1.5, 0.5], 'Voigt')
        np.testing.assert_almost_equal(Cmean_voigt.matrix(), Cmean.matrix())

    def test_uniform(self):
        t = 0.3 * UniformTexture() + 0.5 * UniformTexture()
        np.testing.assert_array_almost_equal(C.Voigt_average().matrix(), (C * t).matrix())

    def test_random_sampling(self):
        w = [0.2, 0.3, 0.5]
        cube = DiscreteTexture.cube()
        goss = DiscreteTexture.Goss()
        brass = DiscreteTexture.brass()
        n = 100000
        tm = w[0] *cube + w[1] * goss + w[2] * brass
        sample = tm.sample(n, seed=123)
        n_cube = np.count_nonzero(sample.angle==cube.orientation.angle)
        assert approx(n_cube, rel=0.02) == w[0] * n
        n_goss = np.count_nonzero(sample.angle == goss.orientation.angle)
        assert approx(n_goss, rel=0.02) == w[1] * n
        n_brass = np.count_nonzero(sample.angle == brass.orientation.angle)
        assert approx(n_brass, rel=0.02) == w[2] * n


if __name__ == '__main__':
    unittest.main()



"""
===========================================================
Plot the Tresca hexagon and the von Mises ellipse
===========================================================

We plot the elastic domain, with respect to the von Mises and Tresca criteria, on one single
Matplotlib axis.

.. note::

    This somehow replicates the figure shown on wikipedia to illustrate the difference between those
    criteria (see `here <https://en.wikipedia.org/wiki/Yield_surface#von_Mises_yield_surface>`_).
"""
from elasticipy.yield_criteria import VonMisesCriterion, TrescaCriterion

sigma_y = 100 # yield stress
mises = VonMisesCriterion(yield_stress=sigma_y)
tresca = TrescaCriterion(yield_stress=sigma_y)
fig, ax = mises.plot2D()
fig, ax = tresca.plot2D(fig=fig, ax=ax, color='blue')
ax.legend()
fig.show()





import numpy as np

a = np.sqrt(2)
KELVIN_MAPPING_MATRIX = np.array([[1, 1, 1, a, a, a],
                                  [1, 1, 1, a, a, a],
                                  [1, 1, 1, a, a, a],
                                  [a, a, a, 2, 2, 2],
                                  [a, a, a, 2, 2, 2],
                                  [a, a, a, 2, 2, 2], ])

VOIGT_MAPPING_MATRIX_COMPLIANCE = [[1, 1, 1, 2, 2, 2],
                                  [1, 1, 1, 2, 2, 2],
                                  [1, 1, 1, 2, 2, 2],
                                  [2, 2, 2, 4, 4, 4],
                                  [2, 2, 2, 4, 4, 4],
                                  [2, 2, 2, 4, 4, 4]]

class MappingConvention:
    """
    Generic class for defining the mapping convention to build a 4th-order tensor from a (6,6) matrix, and the mapping
    convention to use for reciprocal tensor.

    Attributes
    ----------
    matrix : numpy.ndarray
        (6,6) matrix evidencing the coefficient between the 4-index and the 2-index notations
    mapping_inverse : MappingConvention
        Mapping convention to use for the reciprocal tensor
    """
    matrix = np.array(KELVIN_MAPPING_MATRIX)

    @property
    def mapping_inverse(self):
        return self

class KelvinMapping(MappingConvention):
    name = 'Kelvin'

class VoigtMapping(MappingConvention):
    name = 'Voigt'

    def __init__(self, tensor='Stiffness'):
        """
        Create a Voigt mapping convention.

        Parameters
        ----------
        tensor : str
            It can be 'stiffness' or 'compliance'
            Type of tensor we define. Depending on this, the mapping convention will change (see notes).

        Notes
        -----
        For stiffness-like tensors (if ``tensor=stiffness``), the mapping matrix will be:

        .. math::

            \\begin{bmatrix}
                1 & 1 & 1 & 1 & 1 & 1\\\\
                1 & 1 & 1 & 1 & 1 & 1\\\\
                1 & 1 & 1 & 1 & 1 & 1\\\\
                1 & 1 & 1 & 1 & 1 & 1\\\\
                1 & 1 & 1 & 1 & 1 & 1\\\\
                1 & 1 & 1 & 1 & 1 & 1\\\\
            \\end{bmatrix}

        Conversely, for compliance-like tensors (if ``tensor=compliance``), the mapping matrix will be:

        .. math::

            \\begin{bmatrix}
                1 & 1 & 1 & \\sqrt{2} & \\sqrt{2} & \\sqrt{2}\\\\
                1 & 1 & 1 & \\sqrt{2} & \\sqrt{2} & \\sqrt{2}\\\\
                1 & 1 & 1 & \\sqrt{2} & \\sqrt{2} & \\sqrt{2}\\\\
                \\sqrt{2} & \\sqrt{2} & \\sqrt{2} & 2 & 2 & 2\\\\
                \\sqrt{2} & \\sqrt{2} & \\sqrt{2} & 2 & 2 & 2\\\\
                \\sqrt{2} & \\sqrt{2} & \\sqrt{2} & 2 & 2 & 2\\\\
            \\end{bmatrix}
        """
        if tensor == 'Stiffness':
            self.matrix = np.ones((6,6))
            self.tensor_type = 'Stiffness'
        else:
            self.matrix = np.array(VOIGT_MAPPING_MATRIX_COMPLIANCE)
            self.tensor_type = 'Compliance'

    @property
    def mapping_inverse(self):
        if self.tensor_type == 'Stiffness':
            return VoigtMapping(tensor='Compliance')
        else:
            return VoigtMapping(tensor='Stiffness')


from qtpy.QtWidgets import QVBoxLayout, QLabel, QPushButton
from qtpy.QtGui import QPixmap
from qtpy.QtCore import Qt
from importlib import resources

def about(dialog):
    dialog.setWindowTitle("About Elasticipy")
    dialog.setFixedWidth(400)

    layout = QVBoxLayout(dialog)

    logo = QLabel()
    pixmap = QPixmap(
        str(resources.files("elasticipy.resources") / "logo_text.png")
    )
    pixmap = pixmap.scaledToWidth(250, Qt.SmoothTransformation)
    logo.setPixmap(pixmap)
    logo.setAlignment(Qt.AlignCenter)
    layout.addWidget(logo)

    # --- Text ---
    text = QLabel(
        "A Python library for elasticity tensors computations<br><br>"
        "© 2025–2026 Dorian Depriester, MIT Licence"
    )
    text.setAlignment(Qt.AlignCenter)
    layout.addWidget(text)

    # --- Link ---
    link = QLabel(
        '<a href="https://elasticipy.readthedocs.io/">'
        'https://elasticipy.readthedocs.io/</a>'
    )
    link.setAlignment(Qt.AlignCenter)
    link.setOpenExternalLinks(True)
    layout.addWidget(link)

    # --- Bug report ---
    link = QLabel(
        '<a href="https://github.com/DorianDepriester/Elasticipy/issues">'
        'Report a bug</a>'
    )
    link.setAlignment(Qt.AlignCenter)
    link.setOpenExternalLinks(True)
    layout.addWidget(link)

    # --- Close button ---
    close_btn = QPushButton("Close")
    close_btn.clicked.connect(dialog.close)
    layout.addWidget(close_btn)
    return dialog


"""
===========================================================
Compute and plot wave velocities
===========================================================

This example shows how to compute the wave velocities from the stiffness tensor, and how to plot them
"""
###############################################################################
# Define the stiffness tensor for forsterite
# -------------------------------------
#
# We define the stiffness tensor for a forsterite (taken from `the similar tutorial for MTEX users
# <https://mtex-toolbox.github.io/WaveVelocities.html>`_.) as well as the mass density
from elasticipy.tensors.elasticity import StiffnessTensor
C = StiffnessTensor.orthorhombic(phase_name='forsterite',
                                 C11=320, C12=68.2, C13=71.6, C22=196.5, C23=76.8,
                                 C33=233.5, C44=64, C55=77, C66=78.7)
rho = 3.36 # kg/dm³!
print(C)

###############################################################################
# Compute the velocities of primary and secondary waves
# ----------------------------------------------------------
cp, cs_fast, cs_slow = C.wave_velocity(rho)

###############################################################################
# Plot them as three independent pole figures:
# ---------------------------------------------
import matplotlib.pyplot as plt
fig = plt.figure(figsize=(20, 5))
cp.plot_as_pole_figure(subplot_args=(131,), title='p wave', fig=fig)
cs_fast.plot_as_pole_figure(subplot_args=(132,), title='s wave (fast)', fig=fig)
cs_slow.plot_as_pole_figure(subplot_args=(133,), title='s wave (slow)', fig=fig)





"""
==============================================================
Illustrate the normality rule in the principle stresses space
==============================================================
The example shows how to illustrate the normality rule based on the 3D representation the yield surface
"""


###################################################
# Plot the yield surface
#----------------------------------
from elasticipy.yield_criteria import MohrCoulomb
mc = MohrCoulomb(2, -20)
fig=mc.plot3D(xrange=(-6, 6), yrange=(-6, 6), zrange=(-6, 6))

###################################################
# Plot the surface normals
#----------------------------------
from elasticipy.tensors.stress_strain import StressTensor

tensile_stress_x = StressTensor.tensile([1,0,0], 1)
tensile_stress_y = StressTensor.tensile([0,1,0], 1)
biaxial_tension  = tensile_stress_x + tensile_stress_y
tension_compress = tensile_stress_x - tensile_stress_y

fig = mc.draw_surface_normal(fig, tensile_stress_x, color='black', auto_scale=True, label='Tensile x')
fig = mc.draw_surface_normal(fig, tensile_stress_y, color='gray', auto_scale=True, label='Tensile y')
fig = mc.draw_surface_normal(fig, biaxial_tension, color='blue', auto_scale=True, label='Biaxial tension')
mc.draw_surface_normal(fig, tension_compress, color='green', auto_scale=True, label='Tensile/compression')

########################
# .. Notes::
#
#     The `auto_scale=True` option automatically scales the provided stress so that the yield surface is reached.





from paraqus.abaqus import OdbReader
from paraqus.writers import AsciiWriter


ODB_PATH = "C:/TEMP/2Dplate.odb"  # Path to the ODB
MODEL_NAME = "Model-1"  # Can be chosen freely
INSTANCE_NAMES = ["PART-1-1"]  # Which instances will be exported
STEP_NAME = "Step-1"  # Name of the step that will be exported
FRAME_INDEX = -1  # Export the last frame of the step

reader = OdbReader(odb_path=ODB_PATH,
                   model_name=MODEL_NAME,
                   instance_names=INSTANCE_NAMES,
                   )

# Start configuring the reader instance by specifying field outputs and
# node/element groups that will be exported. These must of course be
# available in the output database.

# Field export requests
reader.add_field_export_request("U", field_position="nodes")
reader.add_field_export_request("PEEQ", field_position="elements")
reader.add_field_export_request("PE", field_position="elements")

# Request some element sets, so you can have a closer look at these
# elements
reader.add_set_export_request("ESID", set_type="elements",
                              instance_name="PART-1-1")
reader.add_set_export_request("ETOP", set_type="elements",
                              instance_name="PART-1-1")

# Create a writer that will write the exported results to a .vtu file
vtu_writer = AsciiWriter("vtk_output_billet", clear_output_dir=True)

# The method read_instances loops over all part instances for one
# point in time, and returns ParaqusModel instances for each of them.
# You put them in a list here, so they can be inspected, but it is more
# memory-efficient to use them one after another in a for loop (see
# following tutorials)
instance_models = list(reader.read_instances(step_name=STEP_NAME,
                                             frame_index=FRAME_INDEX))



import unittest
from elasticipy.tensors.fourth_order import FourthOrderTensor, SymmetricFourthOrderTensor
import numpy as np

from elasticipy.tensors.mapping import VoigtMapping
from elasticipy.tensors.second_order import SecondOrderTensor


class TestFourthOrderTensor(unittest.TestCase):
    def test_multidimensionalArrayTensors(self):
        m = 5
        a = np.random.random((m, 6, 6))
        T = FourthOrderTensor(a)
        np.testing.assert_array_almost_equal(a, T._matrix)
        T2 = FourthOrderTensor(T.full_tensor)
        np.testing.assert_array_almost_equal(a, T2._matrix)

    def test_nonsymmetry(self):
        a = np.random.random((3,3,3,3))
        with self.assertRaises(ValueError) as context:
            _ = FourthOrderTensor(a)
        self.assertEqual(str(context.exception), 'The input array does not have minor symmetry')
        T = FourthOrderTensor(a, force_minor_symmetry=True)
        Tfull = T.full_tensor
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    for l in range(3):
                        b = 0.25 * (a[i,j,k,l] + a[j,i,k,l] + a[i,j,l,k] + a[j,i,l,k])
                        np.testing.assert_array_almost_equal(Tfull[i,j,k,l], b)

    def test_inversion(self):
        m = 5
        T = FourthOrderTensor.rand(shape=m)
        assert T._matrix.shape == (m, 6,6)
        Tinv = T.inv()
        TTinv = Tinv.ddot(T)
        eye = FourthOrderTensor.identity(shape=m)
        for i in range(m):
            np.testing.assert_array_almost_equal(TTinv[i].full_tensor, eye[i].full_tensor)

    def test_mult(self):
        m, n, o = 5, 4, 3
        a = FourthOrderTensor.rand(shape=(m,n,o))
        b = 5
        ab = a * b
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_almost_equal(ab[i,j,k]._matrix, a[i,j,k]._matrix * b)

        b = np.random.random((n,o))
        ab = a * b
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_almost_equal(ab[i,j,k]._matrix, a[i,j,k]._matrix * b[j,k])

    def test_zeros_setitem(self):
        m, n = 4, 5
        t = FourthOrderTensor.zeros()
        assert t.shape == ()
        assert np.all(t.full_tensor==0.)

        t = FourthOrderTensor.zeros(n)
        assert t.shape == (n,)

        t = FourthOrderTensor.zeros((m,n))
        assert t.shape == (m, n)

        t[1,3] = np.ones((6,6))
        for i in range(m):
            for j in range(n):
                if (i == 1) and (j == 3):
                    assert np.all(t[i,j] == 1.)
                else:
                    assert np.all(t[i, j] == 0.)

        t0 = t == 0.
        t0_th = np.ones((m, n))
        t0_th[1,3] = 0.
        assert np.all(t0== t0_th)

    def test_div(self):
        m, n, o = 5, 4, 3
        a = FourthOrderTensor.rand(shape=(m,n,o))
        a_div_a = a / a
        np.testing.assert_array_almost_equal(a_div_a.full_tensor, FourthOrderTensor.identity(shape=(m,n,o)).full_tensor)

        half_a = a / 2
        np.testing.assert_array_almost_equal(half_a.full_tensor, a.full_tensor/2)

        b = SecondOrderTensor.rand(shape=(4,3))
        a_div_b = a / b
        np.testing.assert_array_almost_equal(a_div_b.matrix, (a * b.inv()).matrix)

    def test_inconsistent_mapping(self):
        t1 = FourthOrderTensor.rand()
        t2 = FourthOrderTensor.rand(mapping=VoigtMapping())
        t3 = t1 + t2
        assert t3.mapping.name == t1.mapping.name
        np.testing.assert_array_equal(t3._matrix, t1._matrix + t2._matrix)
        t4 = t1.ddot(t2)
        assert t4.mapping.name == t1.mapping.name
        np.testing.assert_array_almost_equal(t4._matrix, np.matmul(t1._matrix, t2._matrix))

    def test_copy(self):
        t1 = FourthOrderTensor.rand()
        t2 = t1.copy()
        assert t2 == t1
        t2.mapping = VoigtMapping()
        assert t2 == t1
        assert not np.all(t1.matrix() == t2.matrix())

    def test_identity(self):
        I = FourthOrderTensor.identity()
        Ifull = I.full_tensor
        eye = np.eye(3)
        a = np.einsum('ik,jl->ijkl', eye, eye)
        b = np.einsum('il,jk->ijkl', eye, eye)
        np.testing.assert_array_equal(Ifull, (a + b) / 2)
        A = FourthOrderTensor.rand()
        IA = I.ddot(A)
        AI = A.ddot(I)
        np.testing.assert_array_almost_equal(A.matrix(), IA.matrix())
        np.testing.assert_array_almost_equal(A.matrix(), AI.matrix())

    def test_identity_spherical_part(self):
        Jfull = FourthOrderTensor.identity_spherical_part().full_tensor
        eye = np.eye(3)
        np.testing.assert_array_equal(Jfull, np.einsum('ij,kl->ijkl',eye, eye) / 3)

    def test_spherical_deviatoric_parts(self):
        A = FourthOrderTensor.rand()
        AJ = A.spherical_part()
        AJJ = AJ.spherical_part()
        np.testing.assert_array_almost_equal(AJ.matrix(), AJJ.matrix())
        AK = A.deviatoric_part()
        AKK = AK.deviatoric_part()
        np.testing.assert_array_almost_equal(AK.matrix(), AKK.matrix())
        AKJ = AK.spherical_part()
        AJK = AJ.deviatoric_part()
        np.testing.assert_array_almost_equal(AKJ.matrix(), np.zeros((6,6)))
        np.testing.assert_array_almost_equal(AJK.matrix(), np.zeros((6, 6)))

    def test_rand(self):
        shapes = [(), (3,2)]
        for shape in shapes:
            t = FourthOrderTensor.rand(shape=shape)
            assert t.shape == shape
            assert np.all(t.full_tensor >=0.)
            assert np.all(t.full_tensor < 1.)

    def test_ones(self):
        A = FourthOrderTensor.ones()
        np.testing.assert_array_almost_equal(A.full_tensor, np.ones((3,3,3,3)))
        Av = FourthOrderTensor.ones(mapping=VoigtMapping())
        np.testing.assert_array_almost_equal(Av.full_tensor, np.ones((3, 3, 3, 3)))
        A2d = FourthOrderTensor.ones(shape=(3,4))
        np.testing.assert_array_almost_equal(A2d.full_tensor, np.ones((3,4,3, 3, 3, 3)))


class TestSymmetricFourthOrderTensor(unittest.TestCase):
    def test_inversion(self):
        m = 5
        T = SymmetricFourthOrderTensor.rand(shape=m)
        Tinv = T.inv()
        TTinv = Tinv.ddot(T)
        eye = SymmetricFourthOrderTensor.identity(shape=m)
        for i in range(m):
            np.testing.assert_array_almost_equal(TTinv[i].full_tensor, eye[i].full_tensor)



if __name__ == '__main__':
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
import os
import sys

from docutils import nodes
from docutils.parsers.rst import Directive
import plotly.io as pio

class PlotlyDirective(Directive):
    has_content = True
    option_spec = {
        'show-code': bool,
    }

    def run(self):
        show_code = self.options.get('show-code', True)
        code = "\n".join(self.content)

        # Exécuter le code pour obtenir la figure
        namespace = {}
        exec(code, namespace)
        fig = namespace.get("fig")
        if fig is None:
            raise RuntimeError("Le bloc .. plotly:: doit définir une variable `fig`.")

        # Générer le HTML de la figure
        html = pio.to_html(fig, include_plotlyjs="cdn", full_html=False)

        # Créer les nœuds pour le code et la figure
        nodes_list = []
        if show_code:
            code_node = nodes.literal_block(code, code)
            code_node['language'] = 'python'
            nodes_list.append(code_node)

        raw_node = nodes.raw('', html, format='html')
        nodes_list.append(raw_node)

        return nodes_list

def setup(app):
    app.add_directive("plotly", PlotlyDirective)

src_path = os.path.abspath('../../src/')
sys.path.insert(0, src_path)
print(f"Chemin ajouté au PYTHONPATH : {sys.path}")

import matplotlib
matplotlib.use("Agg")

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'elasticipy'
copyright = '%Y, Dorian Depriester'
author = 'Dorian Depriester'
release = '4.2.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
   'sphinx_rtd_theme',
   'sphinx.ext.doctest',
   'sphinx.ext.autodoc',
   'sphinx.ext.autosummary',
   'sphinx.ext.napoleon',
   'sphinx.ext.githubpages',
   'sphinx.ext.autosectionlabel',
   'sphinx.ext.mathjax',
   'sphinx.ext.linkcode',
   'sphinx_copybutton',
   'matplotlib.sphinxext.plot_directive',
   'sphinx_gallery.gen_gallery',
]

templates_path = ['_templates']
exclude_patterns = []
copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True
plot_include_source = True
plot_html_show_source_link = False


language = 'english'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['../_static']
html_logo = "logo/logo_text_whitebg.png"
html_favicon = 'logo/favicon.png'
numpydoc_class_members_toctree = False
autoclass_content = 'both'

sphinx_gallery_conf = {
    'examples_dirs': ['examples'],  # Dossier source des exemples
    'gallery_dirs': ['auto_examples'],  # Dossier de sortie
    'filename_pattern': r'\.py',  # Inclut tous les fichiers .py
    'ignore_pattern': r'_template\.py',  # Exclut les fichiers template
    'plot_gallery': True,  # Affiche les figures
    'reference_url': {
        'elasticipy': None,  # Désactive les liens vers la doc API si non configuré
    },
    'capture_repr': ('_repr_html_', '__repr__'),  # Capture les sorties textuelles
    'show_memory': False,  # Désactive l'affichage de la mémoire (optionnel)
}
pio.renderers.default = 'sphinx_gallery'


def linkcode_resolve(domain, info):
    if domain != 'py':
        return None
    if not info['module']:
        return None
    filename = info['module'].replace('.', '/')
    return f"https://github.com/DorianDepriester/Elasticipy/blob/main/src/{filename}.py"



from copy import deepcopy

from matplotlib import pyplot as plt
from orix.quaternion import Orientation
from orix.vector import Vector3d, Miller
from orix.crystal_map import Phase
from scipy.integrate import quad_vec
import numpy as np
from elasticipy.polefigure import add_polefigure
from elasticipy.tensors.fourth_order import FourthOrderTensor
from abc import ABC
from scipy.spatial.transform.rotation import Rotation
from scipy import __version__ as scipy_version

ANGLE_35 = 35.26438968
ANGLE_37 = 36.6992252
ANGLE_54 = 54.73561032
ANGLE_59 = 58.97991646
ANGLE_63 = 63.43494882
ANGLE_74 = 74.20683095

class CrystalTexture(ABC):
    _title = 'Abstract class for crystallographic texture'

    def __init__(self, phase=None):
        self.weight = 1.
        self._details = None
        self.phase = phase
        self.orientation = None

    def mean_tensor(self, tensor):
        """
        Perform the texture-weighted mean of a 4th-order tensor.

        Parameters
        ----------
        tensor : SymmetricFourthOrderTensor
            Reference tensor (unrotated)

        Returns
        -------
        SymmetricFourthOrderTensor
            mean value of the rotated tensor
        """
        pass

    def __mul__(self, other):
        if isinstance(other, FourthOrderTensor):
            return self.mean_tensor(other)
        else:
            t = deepcopy(self)
            t.weight = other
            return t

    def __rmul__(self, other):
        return self * other

    def __add__(self, other):
        # self + other
        if isinstance(other, CrystalTexture):
            return CompositeTexture([self, other])
        elif isinstance(other, CompositeTexture):
            t = deepcopy(other)
            t.texture_list.insert(0, self)
            return t

    def __repr__(self):
        if self._details is None:
            return self._title
        else:
            return self._title + '\n' + self._details

    def _plot_as_pf(self, orientations,
                    uvw=None, UVTW=None, hkl=None, hkil=None,
                    fig=None, projection='lambert', plot_type='plot', ax=None, symmetrise=False, **kwargs):
        miller = Miller(uvw=uvw, hkl=hkl, UVTW=UVTW, hkil=hkil, phase=self.phase)
        if symmetrise:
            miller = miller.symmetrise(unique=True)
        if fig is None:
            fig = plt.figure()
        if ax is None:
            ax = add_polefigure(fig, projection=projection)
        m = orientations.shape[0]
        n = miller.shape[0]
        phi = np.zeros((m, n))
        theta = np.zeros((m, n))
        for i in range(0, n):
            mi = miller[i]
            t = Vector3d(~orientations * mi)
            phi[:, i] = t.azimuth
            theta[:, i] = t.polar
        if plot_type == 'scatter':
            ax.scatter(phi, theta, **kwargs)
        else:
            line, = ax.plot(phi[:, 0], theta[:, 0], **kwargs)
            color = line.get_color()
            ax.plot(phi[:, 1:], theta[:, 1:], color=color)
        ax.set_ylim([0, np.pi / 2])
        return fig, ax

    def plot_as_pole_figure(self, uvw=None, hkl=None, UVTW=None, hkil=None, symmetrise=False, projection='lambert', fig=None, ax=None, **kwargs):
        """
        Plot the pole figure of the crystallographic texture

        Parameters
        ----------
        uvw : list
            Miller indices of directions to plot
        hkl : list
            Miller indices of plane normal to plot
        UVTW : list
            Miller-Bravais indices of directions to plot
        hkil : list
            Miller-Bravais indices of plane normal to plot
        symmetrise : bool, optional
            Symmetrise the direction/plane to plot the family
        projection : str, optional
            Type of projection to use, it can be either stereographic or Lambert
        fig : matplotlib.figure.Figure, optional
            Handle to existing figure, if needed
        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes to plot on
        kwargs
            Keyword arguments to pass to matplotlib's scatter/plot functions

        Returns
        -------
        matplotlib.figure.Figure
            Handle to figure
        matplotlib.projections.polar.PolarAxes
            Axes where the pole figure is plotted

        Examples
        --------
        Plot the <100> pole figure of Goss texture:

        .. plot::

            from elasticipy.crystal_texture import DiscreteTexture

            goss = DiscreteTexture.Goss()
            goss.plot_as_pole_figure(uvw=[1,0,0], symmetrise=True)

        Plot the <110> pole figure of the gamma fibre texture:

        .. plot::

            from elasticipy.crystal_texture import FibreTexture

            gamma = FibreTexture.gamma()
            gamma.plot_as_pole_figure(uvw=[1,1,0], symmetrise=True)
        """
        return self._plot_as_pf(self.orientation,
                                uvw=uvw, hkl=hkl, UVTW=UVTW, hkil=hkil,
                                symmetrise=symmetrise,
                                fig=fig, projection=projection, ax=ax, plot_type='scatter', **kwargs)

    def sample(self, num=50, seed=None):
        """Generate a random sample from the texture component

        Parameters
        ----------
        num : int, optional
            Number of random orientations to generate. Default is 50.
        seed : int, optional
            Seed to use for generating random numbers. Default is None.

        Returns
        -------
        orix.quaternion.Orientation
            Random orientations from the given texture

        Examples
        --------
        Generate 50 orientations from uniform distribution:

        >>> from elasticipy.crystal_texture import UniformTexture
        >>> o = UniformTexture().sample()

        Generate 10 orientations from gamma-fibre

        >>> from elasticipy.crystal_texture import FibreTexture
        >>> o = FibreTexture.gamma().sample(num=10)
        """
        pass

class UniformTexture(CrystalTexture):
    """
    Simple class to define the uniform texture over SO(3)
    """
    _title = 'Uniform texture'

    def __init__(self):
        """
        Create a uniform texture over SO(3)
        """
        super().__init__()
        self._details = 'Uniform distribution over SO(3)'

    def mean_tensor(self, tensor):
        return tensor.infinite_random_average()

    def plot_as_pole_figure(self, miller, projection='lambert', fig=None, ax=None, **kwargs):
        if fig is None:
            fig = plt.figure()
        if ax is None:
            ax = add_polefigure(fig, projection=projection)
        phi = np.linspace(0,2*np.pi)
        theta = np.linspace(0,np.pi)
        phi, theta = np.meshgrid(phi, theta)
        r = np.ones_like(phi)
        ax.contourf(phi, theta, r, **kwargs)
        ax.set_ylim([0, np.pi / 2])
        return fig, ax

    def sample(self, num=50, seed=None):
        if scipy_version >= "1.15.0":
            rand = Rotation.random(num=num, rng=seed)
        else:
            rand = Rotation.random(num=num, random_state=seed)
        return Orientation.from_scipy_rotation(rand)

class DiscreteTexture(CrystalTexture):
    """
    Class to handle classical crystallographic texture.

    Notes
    -----
    This class implements the crystallographic textures listed by [Lohmuller]_

    References
    ----------
    .. [Lohmuller] Lohmuller, P.; Peltier, L.; Hazotte, A.; Zollinger, J.; Laheurte, P.; Fleury, E. Variations of
       the Elastic Properties of the CoCrFeMnNi High Entropy Alloy Deformed by Groove Cold Rolling.
       Materials 2018, 11, 1337. https://doi.org/10.3390/ma11081337
    """
    _title = "Crystallographic texture"

    def __init__(self, orientation, phase=None):
        """
        Create a single-orientation crystallographic texture.

        Parameters
        ----------
        orientation : orix.quaternion.orientation.Orientation
            Orientation of the crystals
        phase : orix.crystal_map.phase_list.Phase
            Phase of the crystal
        """
        super().__init__(phase=phase)
        self.orientation = orientation
        self._details = 'φ1={:.2f}°, ϕ={:.2f}°, φ2={:.2f}°'.format(*orientation.to_euler(degrees=True)[0])

    def mean_tensor(self, tensor):
        return tensor * self.orientation

    @classmethod
    def cube(cls, phase=Phase(point_group='m-3m')):
        """
        Create a Cube crystallographic texture: {100}<100>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        A : Create a A single-orientation crystallographic texture
        brass : Create a Brass single-orientation crystallographic texture
        copper : Create a Copper single-orientation crystallographic texture
        CuT : Create a CuT single-orientation crystallographic texture
        Goss : Create a Goss single-orientation crystallographic texture
        GossBrass : Create a Goss-Brass single-orientation crystallographic texture
        P : Create a P single-orientation crystallographic texture
        S : Create an S single-orientation crystallographic texture
        """
        o = Orientation.from_euler([0, 0, 0], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def Goss(cls, phase=Phase(point_group='m-3m')):
        """
        Create a Goss crystallographic texture: {110}<100>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        A : Create an A single-orientation crystallographic texture
        brass : Create a Brass single-orientation crystallographic texture
        copper : Create a Copper single-orientation crystallographic texture
        cube : Create a Cube single-orientation crystallographic texture
        CuT : Create a CuT single-orientation crystallographic texture
        GossBrass : Create a Goss-Brass single-orientation crystallographic texture
        P : Create a P single-orientation crystallographic texture
        S : Create an S single-orientation crystallographic texture
        """
        o = Orientation.from_euler([0, 45, 0], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def brass(cls, phase=Phase(point_group='m-3m')):
        """
        Create a Brass crystallographic texture: {110}<112>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        copper : Create a Copper single-orientation crystallographic texture
        cube : Create a Cube single-orientation crystallographic texture
        GossBrass : Create a Goss-Brass single-orientation crystallographic texture
        """
        o = Orientation.from_euler([ANGLE_35, 45, 0], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def GossBrass(cls, phase=Phase(point_group='m-3m')):
        """
        Create a Goss/Brass crystallographic texture: {110}<115>

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        brass : Create a Brass single-orientation crystallographic texture
        copper : Create a Copper single-orientation crystallographic texture
        Goss : Create a Goss single-orientation crystallographic texture
        """
        o = Orientation.from_euler([ANGLE_74, 90, 45], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def copper(cls, phase=Phase(point_group='m-3m')):
        """
        Create a copper crystallographic texture: {112}<111>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        brass : Create a Brass single-orientation crystallographic texture
        Goss : Create a Goss single-orientation crystallographic texture
        GossBrass : Create a Goss-Brass single-orientation crystallographic texture
        """
        o = Orientation.from_euler([90, ANGLE_35, 45], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def A(cls, phase=Phase(point_group='m-3m')):
        """
        Create an "A" crystallographic texture: {110}<111>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        cube : Create a Cube single-orientation crystallographic texture
        P : Create a P single-orientation crystallographic texture
        S : Create an S single-orientation crystallographic texture
        """
        o = Orientation.from_euler([ANGLE_35, 90, 45], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def P(cls, phase=Phase(point_group='m-3m')):
        """
        Create a "P"" crystallographic texture: {011}<211>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        A : Create an A single-orientation crystallographic texture
        cube : Create a Cube single-orientation crystallographic texture
        S : Create an S single-orientation crystallographic texture
        """
        o = Orientation.from_euler([ANGLE_54, 90, 45], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def CuT(cls, phase=Phase(point_group='m-3m')):
        """
        Create a CuT crystallographic texture: {552}<115>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        A : Create an A single-orientation crystallographic texture
        P : Create a P single-orientation crystallographic texture
        S : Create an S single-orientation crystallographic texture
        """
        o = Orientation.from_euler([90, ANGLE_74, 45], degrees=True)
        return DiscreteTexture(o, phase=phase)

    @classmethod
    def S(cls, phase=Phase(point_group='m-3m')):
        """
        Create an "S" crystallographic texture: {123}<634>

        Parameters
        ----------
        phase : Phase, optional
            Structure of the crystal. By default, cubic (point group: m-3m) is used.

        Returns
        -------
        DiscreteTexture

        See Also
        --------
        A : Create an A single-orientation crystallographic texture
        CuT : Create a CuT single-orientation crystallographic texture
        P : Create a P single-orientation crystallographic texture
        """
        o = Orientation.from_euler([ANGLE_59, ANGLE_37, ANGLE_63], degrees=True)
        return DiscreteTexture(o, phase=phase)

    def sample(self, num=50, seed=None):
        return Orientation(np.repeat(self.orientation.data, num, axis=0))

class FibreTexture(CrystalTexture):
    _title = 'Fibre texture'

    def __init__(self, orientation, axis, phase=None):
        """
        Create a fibre-type crystallographic texture

        Parameters
        ----------
        orientation : orix.quaternion.orientation.Orientation
            Reference orientation
        axis : list or tuple or numpy.ndarray or orix.vector.Vector3D
            Axis of rotation (in sample CS)
        point_group : orix.phase.point_group.PointGroup, optional
            Point group to use
        """
        super().__init__()
        self.orientation = orientation
        self.axis = Vector3d(axis)
        self.phase = phase

    @classmethod
    def from_Euler(cls, phi1=None, Phi=None, phi2=None, degrees=True, phase=None):
        """
        Create a fibre texture by providing two fixed Bunge-Euler values

        Parameters
        ----------
        phi1 : float
            First Euler angle
        Phi : float
            Second Euler angle
        phi2 : float
            Third Euler angle
        degrees : boolean, optional
            If true (default), the angles must be passed in degrees (in radians otherwise)

        Returns
        -------
        FibreTexture

        See Also
        --------
        from_Miller_axis : Define a fibre texture by aligning a miller direction with a given axis

        Examples
        --------
        A fibre texture corresponding to constant (e.g. zero) values for phi1 and phi2, and uniform distribution of Phi
        on [0,2π[, can be defined as follows:

        >>> from elasticipy.crystal_texture import FibreTexture
        >>> t1 = FibreTexture.from_Euler(phi1=0., phi2=0.)
        >>> t1
        Fibre texture
        φ1= 0.0°, φ2= 0.0°

        Similarly, the following returns a fibre texture for phi1=0 and Phi=0, and uniform distribution of phi2 on
        [0,2π[:

        >>> t2 = FibreTexture.from_Euler(phi1=0., Phi=0.)
        >>> t2
        Fibre texture
        φ1= 0.0°, ϕ= 0.0°
        """
        if phi1 is None:
            orient1 = Orientation.from_euler([0., Phi, phi2] , degrees=degrees)
            orient2 = Orientation.from_euler([1., Phi, phi2] , degrees=degrees)
            angle_list = {'ϕ':Phi, 'φ2':phi2}
        elif Phi is None:
            orient1 = Orientation.from_euler([phi1, 0., phi2], degrees=degrees)
            orient2 = Orientation.from_euler([phi1, 1., phi2], degrees=degrees)
            angle_list = {'φ1':phi1, 'φ2':phi2}
        elif phi2 is None:
            orient1 = Orientation.from_euler([phi1, Phi, 0.] , degrees=degrees)
            orient2 = Orientation.from_euler([phi1, Phi, 1.] , degrees=degrees)
            angle_list = {'φ1':phi1, 'ϕ':Phi}
        else:
            raise ValueError("Exactly two Euler angles are required.")
        axis = (~orient1 * orient2).axis
        a = cls(orient2, axis, phase=phase)
        (k1, v1), (k2, v2) = angle_list.items()
        if not degrees:
            v1 = v1 * 180 / np.pi
            v2 = v2 * 180 / np.pi
        a._details = f"{k1}= {v1}°, {k2}= {v2}°"
        return a

    @classmethod
    def from_Miller_axis(cls, miller, axis):
        """
        Create a perfect fibre crystallographic texture

        Parameters
        ----------
        miller : orix.vector.miller.Miller
            Crystal plane or direction to align with the axis
        axis : tuple or list
            Axis (in sample CS) to align with

        Returns
        -------
        FibreTexture

        See Also
        --------
        from_Euler : define a fibre texture from two Euler angles

        Examples
        --------
        Let's consider a cubic poly-crystal (point group: m-3m), whose orientations are defined by a perfect alignment
        of direction <100> with the Z axis of a sample (therefore a uniform distribution around the Z axis). This
        texture can be defined as:

        >>> from orix.crystal_map import Phase
        >>> from orix.vector.miller import Miller
        >>> from elasticipy.crystal_texture import FibreTexture
        >>> phase = Phase(point_group='m-3m')
        >>> m = Miller(uvw=[1,0,0], phase=phase)
        >>> t = FibreTexture.from_Miller_axis(m, [0,0,1])
        >>> t
        Fibre texture
        <1. 0. 0.> || [0, 0, 1]
        """
        ref_orient = Orientation.from_align_vectors(miller, Vector3d(axis))
        a = cls(ref_orient, axis, phase=miller.phase)
        if miller.coordinate_format == 'uvw' or miller.coordinate_format == 'UVTW':
            miller_str = str(miller.uvw[0])
            miller_str = miller_str.replace('[', '<').replace(']', '>')
        else:
            miller_str = str(miller.hkl[0])
        row_0 = "{miller} || {axis}".format(miller=miller_str, axis=axis)
        a._details = row_0
        return a

    def mean_tensor(self, tensor):
        tensor_ref_orient = tensor * ~self.orientation
        def fun(theta):
            rotation = ~Orientation.from_axes_angles(self.axis, theta)
            tensor_rotated = tensor_ref_orient * rotation
            return tensor_rotated.to_Kelvin()
        circle = 2 * np.pi
        res, *_ = quad_vec(fun, 0, circle)
        return tensor.__class__.from_Kelvin(res / circle)

    def plot_as_pole_figure(self, uvw=None, hkl=None, UVTW=None, hkil=None, projection='lambert', fig=None, ax=None,
                            n_orientations=100, symmetrise=False, **kwargs):
        theta = np.linspace(0, 2 * np.pi, n_orientations)
        orientations = self.orientation * Orientation.from_axes_angles(self.axis, theta)
        return self._plot_as_pf(orientations,
                                uvw=uvw, hkl=hkl, UVTW=UVTW, hkil=hkil,
                                symmetrise=symmetrise,
                                fig=fig, ax=ax, projection=projection, **kwargs)

    def sample(self, num=50, seed=None):
        rng = np.random.default_rng(seed)
        theta = rng.uniform(0.0, 2.0 * np.pi, size=num)
        random_rot = Orientation.from_axes_angles(self.axis, theta)
        return self.orientation * random_rot

    @classmethod
    def gamma(cls):
        """
        Create a gamma fibre-texture: <111> || ND

        Returns
        -------
        FibreTexture

        See Also
        --------
        alpha : create an alpha fibre texture
        epsilon : create an epsilon fibre texture
        """
        phase = Phase(point_group='m3m')
        m = Miller(uvw=[1,1,1], phase=phase)
        return FibreTexture.from_Miller_axis(m, [0,0,1])

    @classmethod
    def alpha(cls):
        """
        Create an alpha fibre-texture: <110> || RD

        Returns
        -------
        FibreTexture

        See Also
        --------
        gamma : create an gamma fibre texture
        epsilon : create an epsilon fibre texture
        """
        phase = Phase(point_group='m3m')
        m = Miller(uvw=[1,1,0], phase=phase)
        return FibreTexture.from_Miller_axis(m, [1,0,0])

    @classmethod
    def epsilon(cls):
        """
        Create an epsilon fibre-texture: <110> || TD

        Returns
        -------
        FibreTexture

        See Also
        --------
        gamma : create an gamma fibre texture
        alpha : create an alpha fibre texture
        """
        phase = Phase(point_group='m3m')
        m = Miller(uvw=[1,1,0], phase=phase)
        return FibreTexture.from_Miller_axis(m, [0,1,0])

class CompositeTexture:
    def __init__(self, texture_list):
        """
        Create a mix of crystal textures

        Parameters
        ----------
        texture_list : list of CrystalTexture
            List of crystal textures to mix
        """
        self.texture_list = list(texture_list)

    def __mul__(self, other):
        # self * other
        if isinstance(other, (float, int)):
            tm = deepcopy(self)
            for t in tm.texture_list:
                t.weight *= other
            return tm
        elif isinstance(other, FourthOrderTensor):
            return self.mean_tensor(other)

    def __rmul__(self, other):
        # other * self
        return self * other

    def __add__(self, other):
        # self + other
        if isinstance(other, CrystalTexture):
            t = deepcopy(self)
            t.texture_list.append(other)
            return t
        elif isinstance(other, CompositeTexture):
            return CompositeTexture(self.texture_list + other.texture_list)

    def __len__(self):
        return len(self.texture_list)

    def __repr__(self):
        title = 'Mixture of crystallographic textures'
        heading = ' Wgt.  Type      Component'
        sep =     ' ------------------------------------------------------------'
        table = []
        for t in self.texture_list:
            if isinstance(t, DiscreteTexture):
                kind = 'discrete'
            elif isinstance(t, UniformTexture):
                kind = 'uniform '
            else:
                kind = 'fibre   '
            table.append(' {:.2f}  {}  {}'.format(t.weight, kind, t._details))
        return '\n'.join([title, heading, sep] + table)

    def mean_tensor(self, tensor):
        """
        Compute the weighted average of a tensor, considering each texture component separately.

        Parameters
        ----------
        tensor : FourthOrderTensor
            Reference tensor (unrotated)

        Returns
        -------
        FourthOrderTensor

        Examples
        --------
        Let consider a mixture of Goss and fibre tensor (with phi1=0 and phi2=0):

        >>> from elasticipy.crystal_texture import DiscreteTexture, FibreTexture
        >>> from elasticipy.tensors.elasticity import StiffnessTensor
        >>> t = DiscreteTexture.Goss() + FibreTexture.from_Euler(phi1=0.0, phi2=0.0)
        >>> t
        Mixture of crystallographic textures
         Wgt.  Type      Component
         ------------------------------------------------------------
         1.00  discrete  φ1=0.00°, ϕ=45.00°, φ2=0.00°
         1.00  fibre     φ1= 0.0°, φ2= 0.0°

        Then, assume that the stiffness tensor is defined as follows:

        >>> C = StiffnessTensor.cubic(C11=186, C12=134, C44=77) # mp-30

        The ODF-weighted Voigt average can be computed as follows:

        >>> Cvoigt = t.mean_tensor(C)
        >>> Cvoigt
        Stiffness tensor (in Voigt mapping):
        [[186.   134.   134.     0.     0.     0.  ]
         [134.   224.25  95.75   0.     0.     0.  ]
         [134.    95.75 224.25  -0.     0.     0.  ]
         [  0.     0.     0.    25.88   0.     0.  ]
         [  0.     0.     0.     0.    57.75  -0.  ]
         [  0.     0.     0.     0.     0.    57.75]]

        Alternatively, on can directly use the following syntax:

        >>> Cvoigt = C * t
        """
        n = len(self)
        t = tensor.__class__.eye(shape=(n,))
        wgt = []
        for i in range(0, n):
            ti = self.texture_list[i]
            wgt.append(ti.weight)
            t[i] = ti.mean_tensor(tensor)
        return t.tensor_average(weights=wgt)

    def plot_as_pole_figure(self,
                            uvw=None, hkl=None, UVTW=None, hkil=None,
                            fig=None, symmetrise=False, projection='lambert',
                            labels=None, **kwargs):
        """
        Plot the pole figure of the composite texture, given a set of Miller indices

        Parameters
        ----------
        miller : orix.vector.miller.Miller
            Miller indices of directions/planes to plot
        projection : str, optional
            Type of projection to use, it can be either stereographic or Lambert
        fig : matplotlib.figure.Figure, optional
            Handle to existing figure, if needed
        symmetrise : bool
            Whether the symmetrise the miller indices
        labels : list of str or tuple of str, optional
            List of labels to use in the legend for each texture component. If not provided, they are automatically
            inferred from each texture component.
        kwargs
            Keyword arguments to pass to matplotlib's scatter/plot functions

        Returns
        -------
        matplotlib.figure.Figure
            Handle to figure
        matplotlib.projections.polar.PolarAxes
            Axes where the pole figure is plotted

        Examples
        --------
        Create a mixture of gamma and alpha fibre texture, and plot the corresponding [1,0,0] pole figure:

        .. plot::

            from elasticipy.crystal_texture import FibreTexture
            from orix.vector import Miller
            from orix.crystal_map import Phase

            texture = FibreTexture.alpha() + FibreTexture.gamma()
            texture.plot_as_pole_figure(uvw=[1,1,0], symmetrise=True)
        """
        if fig is None:
            fig = plt.figure(tight_layout=True)
        ax = add_polefigure(fig, projection=projection)
        for i, t in enumerate(self.texture_list):
            if labels is None:
                label = t._details
            else:
                label = labels[i]
            t.plot_as_pole_figure(uvw=uvw, hkl=hkl, UVTW=UVTW, hkil=hkil,
                                  fig=fig, projection=projection, ax=ax, symmetrise=symmetrise,
                                  label=label, **kwargs)
        return fig, ax

    def sample(self, num=50, seed=None):
        """
        Generate a random sample of orientations from the composite texture.

        Parameters
        ----------
        num : int, optional
            Number of orientations to generate
        seed : int, optional
            Seed for random number generator

        Returns
        -------
        orix.quaternion.Orientation
            Random orientations from the given texture

        Examples
        --------
        Create a balanced mixture of Goss and Brass texture:

        >>> from elasticipy.crystal_texture import DiscreteTexture
        >>> goss = DiscreteTexture.Goss()
        >>> brass = DiscreteTexture.brass()
        >>> texture = goss + brass
        >>> print(texture)
        Mixture of crystallographic textures
         Wgt.  Type      Component
         ------------------------------------------------------------
         1.00  discrete  φ1=0.00°, ϕ=45.00°, φ2=0.00°
         1.00  discrete  φ1=35.26°, ϕ=45.00°, φ2=0.00°

        Now generate a set of 1000 orientations from this texture:

        >>> o = texture.sample(num=1000, seed=123) # Use seed to ensure reproducibility

        One can check that around 50% of these orientations correspond to that of Goss:

        >>> import numpy as np
        >>> np.count_nonzero((o * ~goss.orientation).angle == 0.)
        530

        whereas all the other orientations correspond to brass:

        >>> np.count_nonzero((o * ~brass.orientation).angle == 0.)
        470
        """
        weights = np.array([w.weight for w in self.texture_list])
        weights = weights / weights.sum()
        rng = np.random.default_rng(seed)
        counts = rng.multinomial(num, weights)
        quat = np.zeros((num,4))
        start_index = 0
        for tex, ni in zip(self.texture_list, counts):
            sub_sample = tex.sample(ni, seed=seed)
            quat[start_index:start_index+ni] = sub_sample.data
            start_index += ni
        return Orientation(quat)



"""
=============================================================
Plot in 2D the Mohr-Coulomb and Drucker-Prager yield surface
=============================================================

The example illustrates in 2D the differences between the Mohr-Coulomb and the Drucker-Prager yield criteria.

"""
from elasticipy.yield_criteria import DruckerPrager, MohrCoulomb
c   = 2
phi = -20
mc = MohrCoulomb(c, phi)

#####################################################################################
# When expressed in terms of cohesion (c) and friction angle (phi), there are three ways to 'fit' the Drucker-Prager
# (DP) yield criterion with that of Mohr-Coulomb
# (see `here <https://en.wikipedia.org/wiki/Drucker%E2%80%93Prager_yield_criterion#Expressions_in_terms_of_cohesion_and_friction_angle>`_):

pg1 = DruckerPrager.from_cohesion_friction_angle(c, phi, fit='inside')
pg2 = DruckerPrager.from_cohesion_friction_angle(c, phi, fit='middle')
pg3 = DruckerPrager.from_cohesion_friction_angle(c, phi, fit='outside')

fig, ax = mc.plot2D()
fig, ax = pg1.plot2D(fig=fig, ax=ax, label='DP (inside)', color='green', alpha=0., linestyle='dashed')
fig, ax = pg2.plot2D(fig=fig, ax=ax, label='DP (middle)', color='blue', alpha=0., linestyle='dotted')
fig, ax = pg3.plot2D(fig=fig, ax=ax, label='DP (outside)', color='black', alpha=0., linestyle='dashdot')
ax.legend()
fig.show()





from matplotlib.colors import Normalize
from matplotlib import pyplot as plt, cm
import plotly.graph_objects as go
import numpy as np

def cart2sph(u):
    x = u[:, :, 0],
    y = u[:, :, 1],
    z = u[:, :, 2],
    phi = np.arctan2(y,x)
    theta = np.arccos(z)
    return np.squeeze(phi), np.squeeze(theta)

def _plot3D_plotly(u, r, fig=None, r_range=None, name='', **kwargs):
    if fig is None:
        fig = go.Figure()
    if r_range is None:
        r_range = (r.min(), r.max())
    xyz = (u.T * r.T).T
    phi, theta = cart2sph(u)
    custom_data = np.stack((r.T, np.degrees(theta.T), np.degrees(phi.T)), axis=-1)
    fig.add_trace(go.Surface(
        x=xyz[:, :, 0], y=xyz[:, :, 1], z=xyz[:, :, 2],
        surfacecolor=r,
        colorscale='Viridis',
        cmin=r_range[0],
        cmax=r_range[1],
        customdata=custom_data,
        name=name,
        hovertemplate=(
            "φ: %{customdata[2]:.2f}°<br>"
            "θ: %{customdata[1]:.2f}°<br>"
            "Value: %{customdata[0]:.2f}<br>"
             ),
        **kwargs
    ))
    fig.update_layout(
        scene=dict(
            xaxis_title='X',
            yaxis_title='Y',
            zaxis_title='Z',
        ),
        coloraxis=dict(
            colorscale='Viridis',
            cmin=r_range[0],
            cmax=r_range[1],
            colorbar=dict(title='Values')
        ),
    )
    return fig

def _plot3D_matplotlib(u, r, fig=None, **kwargs):
    if fig is None:
        fig = plt.figure()
    norm = Normalize(vmin=r.min(), vmax=r.max())
    colors = cm.viridis(norm(r))
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    xyz = (u.T * r.T).T
    ax.plot_surface(xyz[:, :, 0], xyz[:, :, 1], xyz[:, :, 2], facecolors=colors, rstride=1, cstride=1,
                    antialiased=False, **kwargs)
    mappable = cm.ScalarMappable(cmap='viridis', norm=norm)
    mappable.set_array([])
    fig.colorbar(mappable, ax=ax)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    return ax

def _draw_plotly_arrow(fig, point, dir, length=1., color='black', cone_scale=2., label=None):
    x_line = [point[0], point[0] + length * dir[0]]
    y_line = [point[1], point[1] + length * dir[1]]
    z_line = [point[2], point[2] + length * dir[2]]

    fig.add_trace(go.Scatter3d(
        x=x_line,
        y=y_line,
        z=z_line,
        mode='lines',
        line=dict(color=color, width=4),
        name=label
    ))

    x_cone = point[0] + length * dir[0]
    y_cone = point[1] + length * dir[1]
    z_cone = point[2] + length * dir[2]

    fig.add_trace(go.Cone(
        x=[x_cone],
        y=[y_cone],
        z=[z_cone],
        u=[cone_scale * dir[0]],
        v=[cone_scale * dir[1]],
        w=[cone_scale * dir[2]],
        colorscale=[[0, color], [1, color]],
        showscale=False,
        sizemode='absolute',
        sizeref=1.0
    ))

    return fig

def _draw_plotly_isosurface(fig, x, y, z, f, label=None, opacity=1.0, color='black'):
    fig.add_trace(go.Isosurface(
        x=x.flatten(),
        y=y.flatten(),
        z=z.flatten(),
        value=f.flatten(),
        isomin=0,
        isomax=0,
        caps=dict(x_show=False, y_show=False),
        colorscale=[[0, color], [1, color]],
        showscale=False,
        opacity=opacity,
        name=label,
    ))
    fig.update_layout(scene=dict(
        xaxis=dict(
            title=dict(
                text=r'σ₁'
            )
        ),
        yaxis=dict(
            title=dict(
                text=r'σ₂'
            )
        ),
        zaxis=dict(
            title=dict(
                text=r'σ₃'
            )
        ),
    ), )
    return fig


"""
===================================
Average stiffness of fibre texture
===================================

This example shows how to compute the Voigt-Reuss-Hill average of crystals with perfect fibre texture.
"""
###############################################################################
# Define fibre texture
# -------------------------------------
#
# We define the stiffness tensor for BCC austenite using its elastic constants. They are taken from the
# `Materials Project (mp-13) <https://next-gen.materialsproject.org/materials/mp-13>`_.
from elasticipy.tensors.elasticity import StiffnessTensor
C = StiffnessTensor.cubic(C11=274, C12=175, C44=89)

###############################################################################
# Create a gamma fibre texture
# -------------------------------------------------
# The gamma texture correspond to alignment of  <1 1 1> with normal direction.
from elasticipy.crystal_texture import FibreTexture
gamma = FibreTexture.gamma()
print(gamma)

######################################
# Plot the corresponding pole figure
# ------------------------------------
# The pole figure of uvw directions, defined with the help of
# `orix <https://orix.readthedocs.io/en/stable/index.html>`_, as pole figures. Eg.:
fig, ax = gamma.plot_as_pole_figure(uvw=[1,0,0])

###############################################################################
# Compute the Hill average
# --------------------------
# The Voigt-Reuss-Hill average of the polycrystal is:
Chill = C.Hill_average(orientations=gamma)
print(Chill)

###############################################################################
# Check transversely isotropy
# ---------------------------
# The transversely isotropy, resulting from the fibre texture along ND, can be evidenced by looking at the Young
# modulus, which is constant along any direction in XY plane:
Chill.Young_modulus.plot3D()




import numpy as np
from elasticipy.tensors.second_order import SymmetricSecondOrderTensor, rotation_to_matrix, is_orix_rotation, \
    SecondOrderTensor, ALPHABET
from scipy.spatial.transform import Rotation
from copy import deepcopy
from elasticipy.tensors.mapping import KelvinMapping, VoigtMapping

kelvin_mapping = KelvinMapping()

def voigt_indices(i, j):
    """
    Translate the two-index notation to one-index notation

    Parameters
    ----------
    i : int or np.ndarray
        First index
    j : int or np.ndarray
        Second index

    Returns
    -------
    Index in the vector of length 6
    """
    voigt_mat = np.array([[0, 5, 4],
                          [5, 1, 3],
                          [4, 3, 2]])
    return voigt_mat[i, j]


def unvoigt_index(i):
    """
    Translate the one-index notation to two-index notation

    Parameters
    ----------
    i : int or np.ndarray
        Index to translate
    """
    inverse_voigt_mat = np.array([[0, 0],
                                  [1, 1],
                                  [2, 2],
                                  [1, 2],
                                  [0, 2],
                                  [0, 1]])
    return inverse_voigt_mat[i]

def _rotate_tensor(full_tensor, r):
    rot_mat = rotation_to_matrix(r)
    str_ein = '...im,...jn,...ko,...lp,...mnop->...ijkl'
    return np.einsum(str_ein, rot_mat, rot_mat, rot_mat, rot_mat, full_tensor)

def _isotropic_matrix(C11, C12, C44):
    C11 = np.asarray(C11)
    C12 = np.asarray(C12)
    C44 = np.asarray(C44)
    shape = np.broadcast_shapes(C11.shape, C12.shape, C44.shape)
    matrix = np.zeros(shape=shape + (6, 6))
    matrix[..., 0, 0] = C11
    matrix[..., 1, 1] = C11
    matrix[..., 2, 2] = C11
    matrix[..., 0, 1] = matrix[..., 1, 0] = C12
    matrix[..., 0, 2] = matrix[..., 2, 0] = C12
    matrix[..., 1, 2] = matrix[..., 2, 1] = C12
    matrix[..., 3, 3] = C44
    matrix[..., 4, 4] = C44
    matrix[..., 5, 5] = C44
    return matrix

class FourthOrderTensor:
    """
    Template class for manipulating symmetric fourth-order tensors.
    """
    _tensor_name = '4th-order'

    def _array_to_Kelvin(self, matrix):
        return matrix / self.mapping.matrix * kelvin_mapping.matrix

    def __init__(self, M, mapping=kelvin_mapping, check_minor_symmetry=True, force_minor_symmetry=False):
        """
        Construct of Fourth-order tensor with minor symmetry.

        Parameters
        ----------
        M : np.ndarray or FourthOrderTensor
            (...,6,6) matrix corresponding to the stiffness tensor, written using the Voigt notation, or array of shape
            (...,3,3,3,3).
        mapping : MappingConvention, optional
            Mapping convention to translate the (3,3,3,3) array to (6,6) matrix.
        check_minor_symmetry : bool, optional
            If true (default), check that the input array have minor symmetries (see Notes). Only used if an array of
            shape (...,3,3,3,3) is passed.
        force_minor_symmetry :
            Ensure that the tensor displays minor symmetry.

        Notes
        -----
        The minor symmetry is defined so that:

        .. math::

            M_{ijkl}=M_{jikl}=M_{jilk}=M_{ijlk}

        Given a generic 4th-order tensor T, the corresponding matrix with respect to Kelvin convention is:

        .. math::

            T =
            \\begin{bmatrix}
                T_{1111}          & T_{1122}          & T_{1133}            & \\sqrt{2}T_{1123} & \\sqrt{2}T_{1113} & \\sqrt{2}T_{1112}\\\\
                T_{2211}          & T_{2222}          & T_{2233}            & \\sqrt{2}T_{2223} & \\sqrt{2}T_{2213} & \\sqrt{2}T_{2212}\\\\
                T_{3311}          & T_{3322}          & T_{3333}            & \\sqrt{2}T_{3323} & \\sqrt{2}T_{3313} & \\sqrt{2}T_{3312}\\\\
                \\sqrt{2}T_{2311} & \\sqrt{2}T_{2322}   & \\sqrt{2}T_{2333} & 2T_{2323}         & 2T_{2313}         & 2T_{2312}\\\\
                \\sqrt{2}T_{1311} & \\sqrt{2}T_{1322}   & \\sqrt{2}T_{1333} & 2T_{423}          & 2T_{1313}         & 2T_{1312}\\\\
                \\sqrt{2}T_{1211} & \\sqrt{2}T_{1222}   & \\sqrt{2}T_{1233} & 2T_{1223}         & 2T_{1223}         & 2T_{1212}\\\\
            \\end{bmatrix}

        Examples
        --------
        Consider a Fourth-order tensor, whose Kelvin matrix is:

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> import numpy as np
        >>> mat = np.array([[100, 200, 300, 0, 0, 0],
        ...                 [-200, 100, 50, 0, 0, 0],
        ...                 [-300, -50, 100, 0, 0, 0],
        ...                 [0, 0, 0, 150, 0, 0],
        ...                 [0, 0, 0, 0, 150, 0],
        ...                 [0, 0, 0, 0, 0, 150]])
        >>> T = FourthOrderTensor(mat)
        >>> print(T)
        4th-order tensor (in Kelvin mapping):
        [[ 100.  200.  300.    0.    0.    0.]
         [-200.  100.   50.    0.    0.    0.]
         [-300.  -50.  100.    0.    0.    0.]
         [   0.    0.    0.  150.    0.    0.]
         [   0.    0.    0.    0.  150.    0.]
         [   0.    0.    0.    0.    0.  150.]]

        If one wants to evaluate the tensor as a (full) (3,3,3,3) array:

        >>> T_array = T.full_tensor

        For instance:

        >>> print(T_array[0,0,0,0])
        100.0

        whereas

        >>> print(T_array[0,1,0,1]) # Corresponds to T_{66}/2
        75.0

        The half factor comes from the Kelvin mapping convention (see Notes). One can also use the Voigt mapping to
        avoid this:

        >>> from elasticipy.tensors.mapping import VoigtMapping
        >>> T_voigt = FourthOrderTensor(mat, mapping=VoigtMapping())
        >>> print(T_voigt)
        4th-order tensor (in Voigt mapping):
        [[ 100.  200.  300.    0.    0.    0.]
         [-200.  100.   50.    0.    0.    0.]
         [-300.  -50.  100.    0.    0.    0.]
         [   0.    0.    0.  150.    0.    0.]
         [   0.    0.    0.    0.  150.    0.]
         [   0.    0.    0.    0.    0.  150.]]

        Although T and T_voigt appear to be the same, note that they are not expressed using the same mapping
        convention. Indeed:

        >>> print(T_voigt.full_tensor[0,0,0,0])
        100.0

        whereas

        >>> print(T_voigt.full_tensor[0,1,0,1])
        150.0

        Alternatively, the differences can be checked with:

        >>> print(T == T_voigt)
        False

        Conversely, let consider the following Voigt matrix:

        >>> mat = np.array([[100, 200, 300, 0, 0, 0],
        ...                 [-200, 100, 50, 0, 0, 0],
        ...                 [-300, -50, 100, 0, 0, 0],
        ...                 [0, 0, 0, 75, 0, 0],
        ...                 [0, 0, 0, 0, 75, 0],
        ...                 [0, 0, 0, 0, 0, 75]])
        >>> T_voigt2 = FourthOrderTensor(mat, mapping=VoigtMapping())
        >>> print(T_voigt2)
        4th-order tensor (in Voigt mapping):
        [[ 100.  200.  300.    0.    0.    0.]
         [-200.  100.   50.    0.    0.    0.]
         [-300.  -50.  100.    0.    0.    0.]
         [   0.    0.    0.   75.    0.    0.]
         [   0.    0.    0.    0.   75.    0.]
         [   0.    0.    0.    0.    0.   75.]]

        Although T and T_voigt2 are not written using the same mapping, we can compare them:

        >>> print(T == T_voigt2) # Same tensors, but different mapping
        True

        whereas

        >>> print(T == T_voigt)  # Different tensors, but same mapping
        False

        This property comes from the fact that the comparison is made independently of the underlying mapping convention.
        """
        if isinstance(mapping, str):
            if mapping.lower() == 'voigt':
                mapping = VoigtMapping()
            elif mapping.lower() == 'kelvin':
                mapping = kelvin_mapping
            else:
                raise ValueError('Mapping must be either "voigt" or "kelvin"')
        self.mapping=mapping
        if isinstance(M, FourthOrderTensor):
            self._matrix = M._matrix
        else:
            M = np.asarray(M)
            if M.shape[-2:] == (6, 6):
                matrix = self._array_to_Kelvin(M)
            elif M.shape[-4:] == (3, 3, 3, 3):
                Mijlk = np.swapaxes(M, -1, -2)
                Mjikl = np.swapaxes(M, -3, -4)
                Mjilk = np.swapaxes(Mjikl, -1, -2)
                if force_minor_symmetry:
                    M = 0.25 * (M + Mijlk + Mjikl + Mjilk)
                elif check_minor_symmetry:
                    symmetry = np.all(M == Mijlk) and np.all(M == Mjikl) and np.all(M == Mjilk)
                    if not symmetry:
                        raise ValueError('The input array does not have minor symmetry')
                matrix = self._full_to_matrix(M)
            else:
                raise ValueError('The input matrix must of shape (...,6,6) or (...,3,3,3,3)')
            self._matrix = matrix
        for i in range(0, 6):
            for j in range(0, 6):
                def getter(obj, I=i, J=j):
                    new_matrix = obj._matrix / kelvin_mapping.matrix * self.mapping.matrix
                    return new_matrix[...,I, J]

                getter.__doc__ = f"Returns the ({i + 1},{j + 1}) component of the {self._tensor_name} matrix."
                component_name = 'C{}{}'.format(i + 1, j + 1)
                setattr(self.__class__, component_name, property(getter))  # Dynamically create the property

    def __repr__(self):
        if (self.ndim == 0) or ((self.ndim==1) and self.shape[0]<5):
            msg = '{} tensor (in {} mapping):\n'.format(self._tensor_name, self.mapping.name)
            matrix = self.matrix(self.mapping)
            with np.printoptions(precision=2, suppress=True):
                msg += matrix.__str__()
        else:
            msg = '{} tensor array of shape {}'.format(self._tensor_name, self.shape)
        return msg

    @property
    def shape(self):
        """
        Return the shape of the tensor array

        Returns
        -------
        tuple
            Shape of the tensor array
        """
        *shape, _, _ = self._matrix.shape
        return tuple(shape)

    @property
    def full_tensor(self):
        """
        Returns the full (unvoigted) tensor as a (3, 3, 3, 3) or (..., 3, 3, 3, 3) array

        Returns
        -------
        np.ndarray
            Full tensor (4-index notation)

        Examples
        --------
        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> I = FourthOrderTensor.eye() # 4th order identity tensor
        >>> print(I)
        4th-order tensor (in Kelvin mapping):
        [[1. 0. 0. 0. 0. 0.]
         [0. 1. 0. 0. 0. 0.]
         [0. 0. 1. 0. 0. 0.]
         [0. 0. 0. 1. 0. 0.]
         [0. 0. 0. 0. 1. 0.]
         [0. 0. 0. 0. 0. 1.]]

        >>> I_full = I.full_tensor
        >>> type(I_full)
        <class 'numpy.ndarray'>
        >>> I_full.shape
        (3, 3, 3, 3)

        When working on tensor arrays, the shape of the resulting numpy array will change accordlingly. E.g.:

        >>> I_array = FourthOrderTensor.eye(shape=(5,6)) # Array of 4th order identity tensor
        >>> I_array.full_tensor.shape
        (5, 6, 3, 3, 3, 3)
        """
        i, j, k, ell = np.indices((3, 3, 3, 3))
        ij = voigt_indices(i, j)
        kl = voigt_indices(k, ell)
        matrix = self._matrix / kelvin_mapping.matrix
        m = matrix[..., ij, kl]
        return m

    def flatten(self):
        """
        Flatten the tensor

        If the tensor array is of shape (m,n,o...,r), the flattened array will be of shape (m*n*o*...*r,).

        Returns
        -------
        SymmetricFourthOrderTensor
            Flattened tensor

        Examples
        --------
        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> T = FourthOrderTensor.rand(shape=(5,6))
        >>> T
        4th-order tensor array of shape (5, 6)
        >>> T.flatten()
        4th-order tensor array of shape (30,)
        """
        shape = self.shape
        if shape:
            t2 = deepcopy(self)
            p = (np.prod(self.shape), 6, 6)
            t2._matrix = self._matrix.reshape(p)
            return t2
        else:
            return self

    def _full_to_matrix(self, full_tensor):
        kl, ij = np.indices((6, 6))
        i, j = unvoigt_index(ij).T
        k, ell = unvoigt_index(kl).T
        return full_tensor[..., i, j, k, ell] * kelvin_mapping.matrix[ij, kl]

    def rotate(self, rotation):
        """
        Apply a single rotation to a tensor, and return its component into the rotated frame.

        Parameters
        ----------
        rotation : Rotation or orix.quaternion.rotation.Rotation
            Rotation to apply

        Returns
        -------
        SymmetricFourthOrderTensor
            Rotated tensor

        Examples
        --------
        Let start from a given tensor, (say ones):

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> T = FourthOrderTensor.ones()
        >>> T
        4th-order tensor (in Kelvin mapping):
        [[1.   1.   1.   1.41 1.41 1.41]
         [1.   1.   1.   1.41 1.41 1.41]
         [1.   1.   1.   1.41 1.41 1.41]
         [1.41 1.41 1.41 2.   2.   2.  ]
         [1.41 1.41 1.41 2.   2.   2.  ]
         [1.41 1.41 1.41 2.   2.   2.  ]]

        Define a rotation. E.g.:

        >>> from scipy.spatial.transform import Rotation
        >>> g = Rotation.from_euler('X', 90, degrees=True)

        Then , apply rotation:

        >>> Trotated = T.rotate(g)
        >>> Trotated
        4th-order tensor (in Kelvin mapping):
        [[ 1.    1.    1.   -1.41  1.41 -1.41]
         [ 1.    1.    1.   -1.41  1.41 -1.41]
         [ 1.    1.    1.   -1.41  1.41 -1.41]
         [-1.41 -1.41 -1.41  2.   -2.    2.  ]
         [ 1.41  1.41  1.41 -2.    2.   -2.  ]
         [-1.41 -1.41 -1.41  2.   -2.    2.  ]]

        Actually, a more simple syntax is:

        >>> T * g
        4th-order tensor (in Kelvin mapping):
        [[ 1.    1.    1.   -1.41  1.41 -1.41]
         [ 1.    1.    1.   -1.41  1.41 -1.41]
         [ 1.    1.    1.   -1.41  1.41 -1.41]
         [-1.41 -1.41 -1.41  2.   -2.    2.  ]
         [ 1.41  1.41  1.41 -2.    2.   -2.  ]
         [-1.41 -1.41 -1.41  2.   -2.    2.  ]]

        Obviously, the original tensor can be retrieved by applying the reverse rotation:

        >>> Trotated * g.inv()
        4th-order tensor (in Kelvin mapping):
        [[1.   1.   1.   1.41 1.41 1.41]
         [1.   1.   1.   1.41 1.41 1.41]
         [1.   1.   1.   1.41 1.41 1.41]
         [1.41 1.41 1.41 2.   2.   2.  ]
         [1.41 1.41 1.41 2.   2.   2.  ]
         [1.41 1.41 1.41 2.   2.   2.  ]]

        If ``g`` is composed of multiple rotations, this will result in a tensor array, corresponding to each rotation:

        >>> import numpy as np
        >>> theta = np.linspace(0, 90, 100)
        >>> g = Rotation.from_euler('X', theta, degrees=True)
        >>> Trotated = T * g
        >>> Trotated
        4th-order tensor array of shape (100,)
        """
        t2 = deepcopy(self)
        rotated_tensor = _rotate_tensor(self.full_tensor, rotation)
        t2._matrix = self._full_to_matrix(rotated_tensor)
        return t2

    @property
    def ndim(self):
        """
        Returns the dimensionality of the tensor (number of dimensions in the orientation array)

        Returns
        -------
        int
            Number of dimensions
        """
        shape = self.shape
        if shape:
            return len(shape)
        else:
            return 0

    def mean(self, axis=None):
        """
        Compute the mean value of the tensor T

        Parameters
        ----------
        axis : int or list of int or tuple of int, optional
            axis along which to compute the mean. If None, the mean is computed on the flattened tensor

        Returns
        -------
        numpy.ndarray
            If no axis is given, the result will be of shape (3,3,3,3).

        Examples
        --------
        Create a random tensor array of shape (5,6):

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> T = FourthOrderTensor.rand(shape=(5,6))
        >>> Overall_mean = T.mean()
        >>> Overall_mean.shape
        ()
        >>> Overall_mean # doctest: +SKIP
        4th-order tensor (in Kelvin mapping):
        [[0.514295   0.52259217 0.42899181 0.77148692 0.64073221 0.73211491]
         [0.49422678 0.43718365 0.40786118 0.8170971  0.68435571 0.67262655]
         [0.48753674 0.51142541 0.44650454 0.76310921 0.67724973 0.69430165]
         [0.53946846 0.75101474 0.73578098 1.04338905 1.21598419 0.99489014]
         [0.75354555 0.61193555 0.82341479 1.11197826 0.89183143 1.20986243]
         [0.66078807 0.70126535 0.63719147 0.87567139 1.05671229 1.03004098]]

         >>> axis_0_mean = T.mean(axis=0)
         >>> axis_0_mean.shape
         (6,)
         >>> axis_1_mean = T.mean(axis=1)
         >>> axis_1_mean.shape
         (5,)
        """
        t2 = deepcopy(self)
        if axis is None:
            axis = tuple([i for i in range(0,self.ndim)])
        t2._matrix = np.mean(self._matrix, axis=axis)
        return t2

    def __add__(self, other):
        new_tensor = deepcopy(self)
        if isinstance(other, np.ndarray):
            if other.shape[-2:] == (6, 6):
                mat = self._matrix + self._array_to_Kelvin(other)
            elif other.shape == (3, 3, 3, 3):
                mat = self._full_to_matrix(self.full_tensor + other)
            else:
                raise ValueError('The input argument must be either a 6x6 matrix or a (3,3,3,3) array.')
        elif isinstance(other, FourthOrderTensor):
            if type(other) == type(self):
                mat = self._matrix + other._matrix
            else:
                raise ValueError('The two tensors to add must be of the same class.')
        else:
            raise ValueError('I don''t know how to add {} with {}.'.format(type(self), type(other)))
        new_tensor._matrix = mat
        return new_tensor

    def __sub__(self, other):
        return self.__add__(-other)

    def __neg__(self):
        t = deepcopy(self)
        t._matrix = -t._matrix
        return t

    def ddot(self, other, mode='pair'):
        """
        Perform tensor product contracted twice (":") between two fourth-order tensors

        Parameters
        ----------
        other : FourthOrderTensor or SecondOrderTensor
            Right-hand side of ":" symbol
        mode : str, optional
            If mode=="pair", the tensors must be broadcastable, and the tensor product are performed on the last axes.
            If mode=="cross", all cross-combinations are considered.

        Returns
        -------
        FourthOrderTensor
            Result from double-contraction

        Examples
        --------
        First, let consider two random arrays of Fourth-order tensors:

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> T1 = FourthOrderTensor.rand(shape=(2,3))
        >>> T2 = FourthOrderTensor.rand(shape=3)
        >>> T1T2_pair = T1.ddot(T2)
        >>> T1T2_pair
        4th-order tensor array of shape (2, 3)

        whereas:

        >>> T1T2_cross = T1.ddot(T2, mode='cross')
        >>> T1T2_cross
        4th-order tensor array of shape (2, 3, 3)

        The command above is equivalent (but way faster) to:

        >>> T1T2_cross_loop = FourthOrderTensor.zeros(shape=(2,3,3))
        >>> for i in range(2):
        ...     for j in range(3):
        ...         for k in range(3):
        ...             T1T2_cross_loop[i,j,k] = T1[i,j].ddot(T2[k])

        One can check that the results are consistent with:

        >>> T1T2_cross_loop == T1T2_cross
        array([[[ True,  True,  True],
                [ True,  True,  True],
                [ True,  True,  True]],
        <BLANKLINE>
               [[ True,  True,  True],
                [ True,  True,  True],
                [ True,  True,  True]]])

        """
        if isinstance(other, FourthOrderTensor):
            if self.ndim == 0 and other.ndim == 0:
                return FourthOrderTensor(np.einsum('ijmn,nmkl->ijkl', self.full_tensor, other.full_tensor))
            else:
                if mode == 'pair':
                    ein_str = '...ijmn,...nmkl->...ijkl'
                else:
                    ndim_0 = self.ndim
                    ndim_1 = other.ndim
                    indices_0 = ALPHABET[:ndim_0]
                    indices_1 = ALPHABET[:ndim_1].upper()
                    indices_2 = indices_0 + indices_1
                    ein_str = indices_0 + 'wxXY,' + indices_1 + 'YXyz->' + indices_2 + 'wxyz'
                matrix = np.einsum(ein_str, self.full_tensor, other.full_tensor)
                return FourthOrderTensor(matrix)
        elif isinstance(other, SecondOrderTensor):
            if self.ndim == 0 and other.ndim == 0:
                return SymmetricSecondOrderTensor(np.einsum('ijkl,kl->ij', self.full_tensor, other.matrix))
            else:
                if mode == 'pair':
                    ein_str = '...ijkl,...kl->...ij'
                else:
                    ndim_0 = self.ndim
                    ndim_1 = other.ndim
                    indices_0 = ALPHABET[:ndim_0]
                    indices_1 = ALPHABET[:ndim_1].upper()
                    indices_2 = indices_0 + indices_1
                    ein_str = indices_0 + 'wxXY,' + indices_1 + 'XY->' + indices_2 + 'wx'
                matrix = np.einsum(ein_str, self.full_tensor, other.matrix)
                return SecondOrderTensor(matrix)

    @classmethod
    def rand(cls, shape=None, **kwargs):
        """
        Populate a Fourth-order tensor with random values in half-open interval [0.0, 1.0).

        Parameters
        ----------
        shape : tuple or int, optional
            Set the shape of the tensor array. If None, the returned tensor will be single.
        kwargs
            Keyword arguments passed to the Fourth-order tensor constructor.

        Returns
        -------
        FourthOrderTensor
            Fourth-order tensor
        """
        if shape is None:
            shape = (6,6)
        elif isinstance(shape, int):
            shape = (shape, 6, 6)
        else:
            shape = tuple(shape) + (6,6)
        mat = np.random.random_sample(shape)
        t = FourthOrderTensor(mat, **kwargs)
        t._matrix = t._matrix * t.mapping.matrix
        return t

    def __mul__(self, other):
        if isinstance(other, (FourthOrderTensor, SecondOrderTensor)):
            return self.ddot(other)
        elif isinstance(other, np.ndarray):
            shape = other.shape
            if other.shape == self.shape[-len(shape):]:
                matrix = self._matrix * other[...,np.newaxis, np.newaxis]
                return self.__class__(matrix)
            else:
                raise ValueError('The arrays to multiply could not be broadcasted with shapes {} and {}'.format(self.shape, other.shape[:-2]))
        elif isinstance(other, Rotation) or is_orix_rotation(other):
            return self.rotate(other)
        elif isinstance(other, (float, int)):
            new_tensor = deepcopy(self)
            new_tensor._matrix = self._matrix * other
            return new_tensor
        else:
            return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (SecondOrderTensor, FourthOrderTensor)):
            return self * other.inv()
        else:
            return self * (1 / other)


    def transpose_array(self):
        """
        Transpose the orientations of the tensor array

        Returns
        -------
        FourthOrderTensor
            The same tensor, but with transposed axes

        Examples
        --------
        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> A = FourthOrderTensor.rand(shape=(3,4))
        >>> A.transpose_array()
        4th-order tensor array of shape (4, 3)
        """
        ndim = self.ndim
        if ndim==0 or ndim==1:
            return self
        else:
            new_array = deepcopy(self)
            new_axes = tuple(range(ndim))[::-1] + (ndim, ndim + 1)
            new_array._matrix = self._matrix.transpose(new_axes)
            return new_array

    def __rmul__(self, other):
        if isinstance(other, (Rotation, float, int, np.number)) or is_orix_rotation(other):
            return self * other
        else:
            raise NotImplementedError('A fourth order tensor can be left-multiplied by rotations or scalar only.')

    def __eq__(self, other):
        if isinstance(other, FourthOrderTensor):
            return np.all(self._matrix == other._matrix, axis=(-1, -2))
        elif isinstance(other, (float, int)) or (isinstance(other, np.ndarray) and other.shape[-2:] == (6, 6)):
            return np.all(self._matrix == other, axis=(-1, -2))
        else:
            raise NotImplementedError('The element to compare with must be a fourth-order tensor '
                                      'or an array of shape (6,6).')

    def __getitem__(self, item):
        if self.ndim:
            sub_tensor = deepcopy(self)
            sub_tensor._matrix = self._matrix[item]
            return sub_tensor
        else:
            raise IndexError('A single tensor cannot be subindexed')

    def __setitem__(self, index, value):
        if isinstance(value, np.ndarray):
            if value.shape[-2:] == (6,6):
                self._matrix[index] = value / self.mapping.matrix * kelvin_mapping.matrix
            elif value.shape[-4:] == (3,3,3,3):
                submatrix = self._full_to_matrix(value)
                self._matrix[index] = submatrix
            else:
                return ValueError('The R.h.s must be either of shape (...,6,6) or (...,3,3,3,3)')
        elif isinstance(value, FourthOrderTensor):
            self._matrix[index] = value._matrix / value.mapping.matrix * self.mapping.matrix
        else:
            raise NotImplementedError('The r.h.s must be either an ndarray or an object of class {}'.format(self.__class__))


    @classmethod
    def identity(cls, **kwargs):
        """
        Construct the Fourth-order identity tensor.

        This is actually an alias for eye().

        Parameters
        ----------
        kwargs
            Keyword arguments passed to the Fourth-order tensor constructor.

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        eye : Fourth-order identity tensor
        """
        return cls.eye(**kwargs)

    @classmethod
    def _broadcast_matrix(cls, M, shape=None, **kwargs):
        if shape is None:
            new_shape = M.shape
        elif isinstance(shape, int):
            new_shape = (shape,) + M.shape
        else:
            new_shape = shape + M.shape
        M_repeat = np.broadcast_to(M, new_shape)
        t = cls(M_repeat, **kwargs)
        t._matrix = t._matrix * t.mapping.matrix / kelvin_mapping.matrix
        return t

    @classmethod
    def eye(cls, shape=(), **kwargs):
        """
        Create a 4th-order identity tensor.

        See notes for definition.

        Parameters
        ----------
        shape : int or tuple, optional
            Shape of the tensor to create
        mapping : Kelvin mapping, optional
            Mapping convention to use. Must be either Kelvin or Voigt.

        Returns
        -------
        FourthOrderTensor or SymmetricFourthOrderTensor
            Identity tensor

        Notes
        -----

        The Fourth-order identity tensor is defined as:

        .. math::

            I_{ijkl} = \\frac12\\left( \\delta_{ik}\\delta_{jl} + \\delta_{il}\\delta_{jk}\\right)

        Examples
        --------
        Create a (single) identity tensor:

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> I = FourthOrderTensor.eye()
        >>> print(I)
        4th-order tensor (in Kelvin mapping):
        [[1. 0. 0. 0. 0. 0.]
         [0. 1. 0. 0. 0. 0.]
         [0. 0. 1. 0. 0. 0.]
         [0. 0. 0. 1. 0. 0.]
         [0. 0. 0. 0. 1. 0.]
         [0. 0. 0. 0. 0. 1.]]

        Alternatively, one can use another mapping convention, e.g. Voigt:

        >>> from elasticipy.tensors.mapping import VoigtMapping
        >>> Iv = FourthOrderTensor.eye(mapping=VoigtMapping())
        >>> print(Iv)
        4th-order tensor (in Voigt mapping):
        [[1.  0.  0.  0.  0.  0. ]
         [0.  1.  0.  0.  0.  0. ]
         [0.  0.  1.  0.  0.  0. ]
         [0.  0.  0.  0.5 0.  0. ]
         [0.  0.  0.  0.  0.5 0. ]
         [0.  0.  0.  0.  0.  0.5]]

        Still, we have:

        >>> print(I == Iv)
        True

        as they correspond to the same tensor, but expressed as a matrix with different mapping conventions. Indeed,
        one can check that:

        >>> import numpy as np
        >>> np.array_equal(I.full_tensor, Iv.full_tensor)
        True
        """
        return cls._broadcast_matrix(np.eye(6), shape=shape, **kwargs)

    @classmethod
    def ones(cls, shape=None, **kwargs):
        """
        Create a 4th-order tensor full of ones.

        Parameters
        ----------
        shape : int or tuple, optional
            Shape of the tensor to create
        kwargs
            keyword arguments passed to the constructor

        Returns
        -------
        FourthOrderTensor

        Examples
        --------
        >>> tensor_of_ones = FourthOrderTensor.ones()
        >>> tensor_of_ones
        4th-order tensor (in Kelvin mapping):
        [[1.   1.   1.   1.41 1.41 1.41]
         [1.   1.   1.   1.41 1.41 1.41]
         [1.   1.   1.   1.41 1.41 1.41]
         [1.41 1.41 1.41 2.   2.   2.  ]
         [1.41 1.41 1.41 2.   2.   2.  ]
         [1.41 1.41 1.41 2.   2.   2.  ]]

        At first sight, the tensor may appear not full of ones at all, but the representation above uses the Kelvin
        mapping convention. Indeed, one can check that the full tensor is actually full of ones. E.g.:

        >>> tensor_of_ones.full_tensor[0,1,0,2]
        np.float64(1.0)

        Alternatively, the Voigt mapping convention may help figuring it out:

        >>> from elasticipy.tensors.mapping import VoigtMapping
        >>> tensor_of_ones_voigt = FourthOrderTensor.ones(mapping=VoigtMapping())
        >>> tensor_of_ones_voigt
        4th-order tensor (in Voigt mapping):
        [[1. 1. 1. 1. 1. 1.]
         [1. 1. 1. 1. 1. 1.]
         [1. 1. 1. 1. 1. 1.]
         [1. 1. 1. 1. 1. 1.]
         [1. 1. 1. 1. 1. 1.]
         [1. 1. 1. 1. 1. 1.]]

        although both tensors are actually the same:

        >>> print(tensor_of_ones == tensor_of_ones_voigt)
        True
        """
        return cls._broadcast_matrix(kelvin_mapping.matrix, shape=shape, **kwargs)

    @classmethod
    def identity_spherical_part(cls, shape=(), **kwargs):
        """
        Return the spherical part of the identity tensor.

        See Notes for mathematical definition.

        Parameters
        ----------
        shape : tuple of int, optional
            Shape of the tensor to create
        kwargs
            Keyword arguments passed to the Fourth-order tensor constructor.

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        identity_tensor : return the identity tensor
        identity_deviatoric_part : return the deviatoric part of the identity tensor

        Notes
        -----
        The spherical part of the identity tensor is defined as:

        .. math::

            J_{ijkl} = \\frac13 \\delta_{ij}\\delta_{kl}

        Examples
        --------
        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> J = FourthOrderTensor.identity_spherical_part()
        >>> print(J)
        4th-order tensor (in Kelvin mapping):
        [[0.33 0.33 0.33 0.   0.   0.  ]
         [0.33 0.33 0.33 0.   0.   0.  ]
         [0.33 0.33 0.33 0.   0.   0.  ]
         [0.   0.   0.   0.   0.   0.  ]
         [0.   0.   0.   0.   0.   0.  ]
         [0.   0.   0.   0.   0.   0.  ]]

        On can check that J has zero deviatoric part:

        >>> J.deviatoric_part()
        4th-order tensor (in Kelvin mapping):
        [[0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]]
        """
        A = np.zeros((6, 6))
        A[:3, :3] = 1 / 3
        return cls._broadcast_matrix(A, shape=shape, **kwargs)

    @classmethod
    def identity_deviatoric_part(cls, **kwargs):
        """
        Return the deviatoric part of the identity tensor.

        See notes for the mathematical definition.

        Parameters
        ----------
        kwargs
            keyword arguments passed to eye constructor

        Returns
        -------
        FourthOrderTensor or SymmetricTensor

        See Also
        --------
        identity_tensor : return the identity tensor
        identity_spherical_part : return the spherical part of the identity tensor

        Notes
        -----
        The deviatoric part of the identity tensor is defined as:

        .. math::

            K = I - J

        where :math:`I` and :math:`J` denote the identity and the deviatoric part of the identity tensor, respectively.

        Examples
        --------
        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> K = FourthOrderTensor.identity_deviatoric_part()
        >>> print(K)
        4th-order tensor (in Kelvin mapping):
        [[ 0.67 -0.33 -0.33  0.    0.    0.  ]
         [-0.33  0.67 -0.33  0.    0.    0.  ]
         [-0.33 -0.33  0.67  0.    0.    0.  ]
         [ 0.    0.    0.    1.    0.    0.  ]
         [ 0.    0.    0.    0.    1.    0.  ]
         [ 0.    0.    0.    0.    0.    1.  ]]

        One can check that K has zero spherical part:

        >>> print(K.spherical_part())
        4th-order tensor (in Kelvin mapping):
        [[0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]]
        """
        I = FourthOrderTensor.identity(**kwargs)
        J = FourthOrderTensor.identity_spherical_part(**kwargs)
        return I-J

    def spherical_part(self):
        """
        Return the spherical part of the tensor

        Returns
        -------
        FourthOrderTensor
            Spherical part of the tensor

        See Also
        --------
        identity_tensor : return the identity tensor
        deviatoric_part : return the deviatoric part of the tensor
        """
        I = self.identity_spherical_part(shape=self.shape)
        return I.ddot(self)

    def deviatoric_part(self):
        """
        Return the deviatoric part of the tensor

        Returns
        -------
        FourthOrderTensor
            Deviatoric part of the tensor

        See Also
        --------
        identity_tensor : return the identity tensor
        spherical_part : return the spherical part of the tensor
        """
        K = self.identity_deviatoric_part(shape=self.shape)
        return K.ddot(self)

    def inv(self):
        """
        Invert the tensor. The inverted tensors inherits the properties (if any)

        Returns
        -------
        FourthOrderTensor
            Inverse tensor

        Examples
        --------
        Let consider a random Fourth-order tensor:

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> T = FourthOrderTensor.rand()
        >>> print(T) # doctest: +SKIP

        >>> Tinv = T.inv()
        >>> print(Tinv) # doctest: +SKIP

        One can check that ``T.ddot(Tinv)`` and ``Tinv.ddot(T)`` are really close to the identity tensor:

        >>> I = FourthOrderTensor.eye()
        >>> (T.ddot(Tinv) - I) * 1e16 # doctest: +SKIP
        4th-order tensor (in Kelvin mapping):
        [[ 1.00000000e+00  0.00000000e+00  1.11022302e-16  0.00000000e+00
           3.14018492e-16  0.00000000e+00]
         [-2.22044605e-16  1.00000000e+00  2.77555756e-17 -1.25607397e-15
           2.35513869e-16 -7.85046229e-17]
         [ 1.55431223e-15  0.00000000e+00  1.00000000e+00 -3.14018492e-16
           0.00000000e+00 -4.71027738e-16]
         [-6.28036983e-16 -4.39625888e-15  0.00000000e+00  1.00000000e+00
           0.00000000e+00  1.11022302e-16]
         [-1.25607397e-15 -4.39625888e-15  0.00000000e+00  2.55351296e-15
           1.00000000e+00 -1.66533454e-16]
         [-5.88784672e-16 -1.17756934e-15 -2.62499833e-16  5.96744876e-16
           1.24900090e-16  1.00000000e+00]]

        >>> (Tinv.ddot(T) - I) * 1e16 # doctest: +SKIP
        4th-order tensor (in Kelvin mapping):
        [[ 1.00000000e+00 -1.33226763e-15 -3.99680289e-15 -6.90840682e-15
          -7.53644380e-15 -2.51214793e-15]
         [ 2.33146835e-15  1.00000000e+00  1.55431223e-15 -7.85046229e-16
           1.41308321e-15  9.42055475e-16]
         [ 3.88578059e-16  1.11022302e-16  1.00000000e+00 -3.92523115e-16
          -1.57009246e-16  1.57009246e-16]
         [-5.88784672e-17 -1.86448479e-16 -1.47196168e-16  1.00000000e+00
          -2.49800181e-16 -2.08166817e-16]
         [ 5.10280049e-16  7.85046229e-17 -7.85046229e-17  1.27675648e-15
           1.00000000e+00  7.77156117e-16]
         [-7.85046229e-16 -6.28036983e-16  6.28036983e-16  2.44249065e-15
           3.55271368e-15  1.00000000e+00]]

        This function obvisouly also works for tensor arrays. E.g.:

        >>> T = FourthOrderTensor.rand(shape=(5,3))
        >>> Tinv = T.inv()
        >>> Tinv.shape
        (5, 3)

        Again, one can check that ``T.ddot(Tinv)`` is close to the array of identity tensors:

        >>> import numpy as np
        >>> I = FourthOrderTensor.eye(shape=(5,3))
        >>> np.max(T.ddot(Tinv).matrix() - I.matrix())  # doctest: +SKIP
        5.906386491005833e-14
        """
        matrix_inv = np.linalg.inv(self._matrix)
        t = self.__class__(matrix_inv, mapping=kelvin_mapping)
        t.mapping = self.mapping.mapping_inverse
        return t

    @classmethod
    def zeros(cls, shape=(), **kwargs):
        """
        Create a fourth-order tensor populated with zeros

        Parameters
        ----------
        shape : int or tuple, optional
            Shape of the tensor to create
        kwargs
            Keyword arguments passed to the FourthOrderTensor constructor

        Returns
        -------
        FourthOrderTensor

        Examples
        --------
        The single-valued null 4th order tensor is just:

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> FourthOrderTensor.zeros()
        4th-order tensor (in Kelvin mapping):
        [[0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]
         [0. 0. 0. 0. 0. 0.]]

        One can also create an array of such tensors:

        >>> zeros_tensor = FourthOrderTensor.zeros(shape=3)

        and check that it populated with zeros:

        >>> zeros_tensor == 0.
        array([ True,  True,  True])
        """
        if isinstance(shape, int):
            shape = (shape, 6, 6)
        else:
            shape = shape + (6,6)
        zeros = np.zeros(shape)
        return cls(zeros, **kwargs)

    def matrix(self, mapping_convention=None):
        """
        Returns the components of the tensor as a matrix.

        Parameters
        ----------
        mapping_convention : VoigtMapping, optional
            Mapping convention to use for the returned matrix. If not provided, that of the tensor is used.

        Returns
        -------
        numpy.ndarray
            Components of the tensor as a matrix

        Examples
        --------
        Create an identity 4th-order tensor:

        >>> from elasticipy.tensors.fourth_order import FourthOrderTensor
        >>> t = FourthOrderTensor.eye()

        Its matrix with respect to Kelvin mapping is:

        >>> t.matrix()
        array([[1., 0., 0., 0., 0., 0.],
               [0., 1., 0., 0., 0., 0.],
               [0., 0., 1., 0., 0., 0.],
               [0., 0., 0., 1., 0., 0.],
               [0., 0., 0., 0., 1., 0.],
               [0., 0., 0., 0., 0., 1.]])

        whereas, when using the Voigt mapping, we have:

        >>> from elasticipy.tensors.mapping import VoigtMapping
        >>> t.matrix(mapping_convention=VoigtMapping())
        array([[1. , 0. , 0. , 0. , 0. , 0. ],
               [0. , 1. , 0. , 0. , 0. , 0. ],
               [0. , 0. , 1. , 0. , 0. , 0. ],
               [0. , 0. , 0. , 0.5, 0. , 0. ],
               [0. , 0. , 0. , 0. , 0.5, 0. ],
               [0. , 0. , 0. , 0. , 0. , 0.5]])

        For stiffness tensors, the default mapping convention is Voigt, so that:

        >>> from elasticipy.tensors.elasticity import StiffnessTensor, ComplianceTensor
        >>> StiffnessTensor.eye().matrix()
        array([[1. , 0. , 0. , 0. , 0. , 0. ],
               [0. , 1. , 0. , 0. , 0. , 0. ],
               [0. , 0. , 1. , 0. , 0. , 0. ],
               [0. , 0. , 0. , 0.5, 0. , 0. ],
               [0. , 0. , 0. , 0. , 0.5, 0. ],
               [0. , 0. , 0. , 0. , 0. , 0.5]])

        whereas for compliance tensor, the default mapping convention gives:

        >>> ComplianceTensor.eye().matrix()
        array([[1., 0., 0., 0., 0., 0.],
               [0., 1., 0., 0., 0., 0.],
               [0., 0., 1., 0., 0., 0.],
               [0., 0., 0., 2., 0., 0.],
               [0., 0., 0., 0., 2., 0.],
               [0., 0., 0., 0., 0., 2.]])
        """
        matrix = self._matrix
        if mapping_convention is None:
            mapping_convention = self.mapping
        elif isinstance(mapping_convention, str):
            if mapping_convention.lower() == 'voigt':
                mapping_convention = VoigtMapping()
            elif mapping_convention.lower() == 'kelvin':
                mapping_convention = kelvin_mapping
            else:
                raise ValueError('Mapping convention must be either Kelvin or Voigt')
        return matrix / kelvin_mapping.matrix * mapping_convention.matrix

    def copy(self):
        """Create a copy of the tensor"""
        a = deepcopy(self)
        return a

    def tensor_average(self, weights=None, axis=0):
        """
        Compute the average of a tensor array.

        Parameters
        ----------
        weights : list of float or tuple of floats, optional
            If provided, each tensor value is weighted accordingly
        axis : int, optional
            Axis to compute the average over (default: 0)

        Returns
        -------
        FourthOrderTensor
        """
        t = deepcopy(self)
        t._matrix = np.average(t._matrix, weights=weights, axis=axis)
        return t

class SymmetricFourthOrderTensor(FourthOrderTensor):
    _tensor_name = 'Symmetric 4th-order'

    def __init__(self, M, check_symmetries=True, force_symmetries=False, **kwargs):
        """
        Construct a fully symmetric fourth-order tensor from a (...,6,6) or a (...,3,3,3,3) array.

        The input matrix must be symmetric, otherwise an error is thrown (except if ``check_symmetry==False``, see
        below)

        Parameters
        ----------
        M : np.ndarray or FourthOrderTensor
            (6,6) matrix corresponding to the stiffness tensor, or slices of (6,6) matrices or array of shape
            (...,3,3,3,3).
        check_symmetries : bool, optional
            Whether to check or not that the tensor to built displays both major and minor symmetries (see Notes).
        force_symmetries : bool, optional
            If true, ensure that the tensor displays both minor and major symmetries.

        Notes
        -----
        The major symmetry is defined so that:

        .. math::

            M_{ijkl}=M_{klij}

        whereas the minor symmetry is:

        .. math::

            M_{ijkl}=M_{jikl}=M_{jilk}=M_{ijlk}
        """
        super().__init__(M, check_minor_symmetry=check_symmetries, force_minor_symmetry=force_symmetries, **kwargs)
        if force_symmetries:
            self._matrix = 0.5 * (self._matrix + self._matrix.swapaxes(-1, -2))
        elif check_symmetries and not np.all(np.isclose(self._matrix, self._matrix.swapaxes(-1, -2))):
            raise ValueError('The input matrix must be symmetric')

    def linear_invariants(self):
        """
        Compute the linear invariants of the tensor, or tensor array.

        If the object is a tensor array, the linear invariants are returned as arrays of each invariant. See notes for
        the actual definitions.

        Returns
        -------
        A1 : float or np.ndarray
            First linear invariant
        A2 : float or np.ndarray
            Second linear invariant

        See Also
        --------
        quadratic_invariants : compute the quadratic invariants of a fourth-order tensor

        Notes
        -----
        The linear invariants are:

        .. math::

            A_1=C_{ijij}

            A_2=C_{iijj}

        """
        t = self.full_tensor
        A1 = np.einsum('...ijij->...',t)
        A2 = np.einsum('...iijj->...',t)
        return A1, A2

    def quadratic_invariants(self):
        """
        Compute the quadratic invariants of the tensor, or tensor array.

        If the object is a tensor array, the returned values are arrays of each invariant. See notes for definitions.

        Returns
        -------
        B1, B2, B3, B4, B5 : float or np.ndarray

        See Also
        --------
        linear_invariants : compute the linear invariants of a Fourth-order tensor

        Notes
        -----
        The quadratic invariants are defined as [Norris]_:

        .. math::

            B_1 = C_{ijkl}C_{ijkl}

            B_2 = C_{iikl}C_{jjkl}

            B_3 = C_{iikl}C_{jkjl}

            B_4 = C_{kiil}C_{kjjl}

            B_5 = C_{ijkl}C_{ikjl}

        References
        ----------
        .. [Norris] Norris, A. N. (22 May 2007). "Quadratic invariants of elastic moduli". The Quarterly Journal of Mechanics
         and Applied Mathematics. 60 (3): 367–389. doi:10.1093/qjmam/hbm007

        """
        t = self.full_tensor
        B1 = np.einsum('...ijkl,...ijkl->...', t, t)
        B2 = np.einsum('...iikl,...jjkl->...', t, t)
        B3 = np.einsum('...iikl,...jkjl->...', t, t)
        B4 = np.einsum('...kiil,...kjjl->...', t, t)
        B5 = np.einsum('...ijkl,...ikjl->...', t, t)
        return B1, B2, B3, B4, B5

    def infinite_random_average(self):
        """
        Compute the average of the tensor, assuming that an infinite number of random orientations is applied.

        Returns
        -------
        FourthOrderTensor
            Average tensor or tensor array. The returned array will be of the same shape as the input object.
        """
        new_tensor = deepcopy(self)
        matrix = self._matrix / kelvin_mapping.matrix
        A = matrix[..., 0, 0] + matrix[..., 1, 1] + matrix[..., 2, 2]
        B = matrix[..., 0, 1] + matrix[..., 0, 2] + matrix[..., 1, 2]
        C = matrix[..., 3, 3] + matrix[..., 4, 4] + matrix[..., 5, 5]
        C11 = 1/5  * A + 2/15 * B + 4/15 * C
        C12 = 1/15 * A + 4/15 * B - 2/15 * C
        C44 = 1/15 * A - 1/15 * B + 1/5 * C
        new_matrix = _isotropic_matrix(C11, C12, C44)
        new_tensor._matrix = new_matrix * kelvin_mapping.matrix
        return new_tensor

    @classmethod
    def rand(cls, shape=None, **kwargs):
        t1 = super().rand(shape)
        return cls(t1, force_symmetries=True, **kwargs)




"""
=====================================
Average stiffness of discrete texture
=====================================

This example shows how to compute the Voigt average of crystals with a mixture of discrete textures.
"""
###############################################################################
# Define fibre texture
# -------------------------------------
#
# We define the stiffness tensor for BCC austenite using its elastic constants. They are taken from the
# `Materials Project (mp-13) <https://next-gen.materialsproject.org/materials/mp-13>`_.
from elasticipy.tensors.elasticity import StiffnessTensor
C = StiffnessTensor.cubic(C11=274, C12=175, C44=89)

###############################################################################
# Create two discrete textures
# -------------------------------------------------
# The Brass texture corresponds to {110}<112>, whereas the copper texture corresponds to {112}<111>:
from elasticipy.crystal_texture import DiscreteTexture
brass = DiscreteTexture.brass()
copper = DiscreteTexture.copper()
print(brass)
print(copper)

######################################
# Create a composite texture
# ------------------------------------
# We assume that the overall texture consists in 60% brass and 40% copper:
texture = 0.6 * brass + 0.4 * copper
print(texture)

###############################################################################
# Plot the pole figures
# --------------------------
fig, ax = texture.plot_as_pole_figure([1,0,0], symmetrise=True, labels=['brass', 'copper'])
ax.legend()

###############################################################################
# Compute the Voigt average
# --------------------------
# The Voigt average of the polycrystal is:
Cvoigt = C.Voigt_average(orientations=texture)
print(Cvoigt)

###############################################################################
# Plot the Young modulus
# -------------------------------
Cvoigt.Young_modulus.plot3D()


